/*******************************************************************/
/*******************************************************************/
/*                                                                 */
/*         CLASS GROUP AND REGULATOR (McCURLEY, BUCHMANN)          */
/*                   QUADRATIC FIELDS                              */
/*                                                                 */
/*******************************************************************/
/*******************************************************************/
/* $Id: buch1.c,v 1.927.0.13 1997/07/12 17:20:41 karim Exp karim $ */

#include "genpari.h"
static const long CBUCH = 15; /*DOIT ETRE DE LA FORME 2^k-1 !!*/
#define HASHT 1024

typedef struct qf_data
{
  GEN D;
  GEN isqrtD; 
  GEN sqrtD;
  long sens;
  long precreg; /* = 0  iff qf is imaginary */
} qf_data;

static GEN comprealform3(GEN x, GEN y, qf_data *data);
static GEN comprealform5(GEN x, GEN y, qf_data *data);
static GEN initializeform3(long *ex,GEN **tabform,long cardtab,qf_data *data);
static GEN initializeform5(long *ex,GEN **tabform,long cardtab,qf_data *data);
static GEN lfunc(GEN D);
static GEN redrealform(GEN x, qf_data *data);
static GEN redrealform_proto(GEN x, qf_data *data, GEN (red)(GEN,qf_data *));
static GEN rhorealform3(GEN x, qf_data *data);
static GEN rhorealform5(GEN x, qf_data *data);
static long *largeprime(long q1, long *ex, long np, long nrho, long cardsubbase, long **hashtab);
static long factorbasequad(GEN d, long n2, long n, long **ptnum, long **ptbase, long *ptkc, long *badprim, long *nbbadprim);
static long factorisequad(GEN f, long n, long limp, long *ptlo, long *primfact, long *expoprimfact, long *badprim, long nbbadprim, long *base, long limhash);

#define redrealform3(x,data) redrealform_proto((x),(data),rhorealform3)
#define redrealform5(x,data) redrealform_proto((x),(data),rhorealform5)

double
check_bach(double cbach, double B)
{
  if (cbach > B - 0.01)
   err(talker,"sorry, buchxxx couldn't deal with this field PLEASE REPORT!");
  return min(2*cbach,B);
}

static GEN
gcdrealnoer(GEN a,GEN b,long *pte)
{
  long av,tetpil,e;
  GEN k1,r;

  if (typ(a)==t_INT)
  {
    if (typ(b)==t_INT) return mppgcd(a,b);
    k1=cgetr(lg(b)); affir(a,k1); a=k1;
  }
  else if (typ(b)==t_INT) { k1=cgetr(lg(a)); affir(b,k1); b=k1; }
  if (expo(a)<-5) return absr(b);
  if (expo(b)<-5) return absr(a);
  av=avma; a=absr(a); b=absr(b);
  while (expo(b) >= -5  && signe(b))
  {
    k1=gcvtoi(divrr(a,b),&e);
    if (e>0) { avma=av; *pte=e; return gzero; }
    r=subrr(a,mulir(k1,b)); a=b; b=r;
  }
  tetpil=avma; *pte=expo(b);
  return gerepile(av,tetpil,gabs(a,0));
}

static void
freehash(long *pt)
{
  long *q;
  while (pt) { q=(long*)*pt; free(pt); pt=q; }
}

static long *
subfactorbase(GEN *w, double ll, long kc, long* vectprime,
                  long* vperm, long *ptnbram, qf_data *data)
{
  long i,j,k,nbidp,pp,*subbase,pro[100];
  double prod;
  GEN p1;

  i=0; *ptnbram=0; prod=1;
  if (ll<=1.1) ll=1.1;
  for (j=1; j<=kc && prod<=ll; j++)
  {
    pp=vectprime[j];
    if (pp>0) { pro[++i]=pp; prod*=pp; vperm[i]=j; }
    else (*ptnbram)++;
  }
  if (prod<=ll) { *ptnbram= -1; return NULL; }
  nbidp=i;
  for (k=1; k<j; k++)
    if (vectprime[k]<=0) vperm[++i]=k;
  *w=cgetg(nbidp+1,t_COL);

  if (data->precreg) /* real */
    for (j=1; j<=nbidp; j++)
    {
      p1=primeform(data->D,stoi(pro[j]),data->precreg);
      (*w)[j] = (long) redrealform(p1,data);
    }
  else
    for (j=1; j<=nbidp; j++) /* imaginary */
    {
      p1=primeform(data->D,stoi(pro[j]),0);
      (*w)[j]=(long)p1;
    }

  subbase = (long*) xmalloc(sizeof(long)*(nbidp+1));
  subbase[0]=nbidp;
  for (j=1; j<=nbidp; j++) subbase[j]=pro[j];
  return subbase;
}

static GEN **
powsubfact(GEN w, long n, long a, GEN unform, qf_data *data)
{
  GEN **x = (GEN**) xmalloc(sizeof(GEN*)*(n+1));
  long i,j;

  for (i=1; i<=n; i++)
    x[i] = (GEN*) xmalloc(sizeof(GEN)*(a+1));

  if (data->precreg) /* real */
    for (i=1; i<=n; i++)
    {
      x[i][0] = gcopy(unform);
      for (j=1; j<=a; j++)
	x[i][j]=comprealform5(x[i][j-1],(GEN)w[i],data);
    }
  else /* imaginary */
    for (i=1; i<=n; i++)
    {
      x[i][0]=gcopy(unform);
      for (j=1; j<=a; j++)
	x[i][j]=compimag(x[i][j-1],(GEN)w[i]);
    }

  if (DEBUGLEVEL)
  { 
    fprintferr("temps powsubfact: ");
    fprintferr("%ld\n",timer2()); flusherr();
  }
  return x;
}

GEN
buchimag(GEN D, GEN gcbach, GEN gcbach2, GEN gCO)
{
  long CO;
  double cbach,cbach2;
  long limc,limc2,mglob,m,cp,nbram,auxrel,lo,lo1,nlze,col;
  long *fpd,b1,b2,fpc,pp,ep,b,extrarel,c,limhash;
  long kc2,*numbase,*base,*subbase;
  long **mat,**matinit,*vectprime,*vperm,*vinvperm;
  long primfact[100],expoprimfact[100],primfact1[100],expoprimfact1[100];
  long badprim[100],nbbadprim;
  long av=avma,tetpil,kc,kcco,kccopro,i,j,*pro,p2,*p1,*ex,q,s,nbtest,mm,av3;
  long sizeofmit,k,nrelsup,nreldep;
  double drc,lim,logd;
  GEN unform,dr,v,u1u2,u1,c_1;
  GEN matc,matalpha,**vp,form,form1,pc,mit,met,mot,p3,p4,basecl;
  GEN extramat,extramatc,pdep;
  long* hashtab[HASHT];
  qf_data *DATA;

  if (DEBUGLEVEL) timer2();
  if (typ(D)!=t_INT || typ(gCO)!=t_INT) err(typeer,"buchimag");
  if (!signe(D)) err(bucher2);
  if (signe(D) > 0)
    err(talker,"positive discriminant in buchimag, try buchreal");
  s=mod4(D); if (s==1 || s==2) err(funder2,"buchimag");
  CO=itos(gCO);
  cbach=gtodouble(gcbach); cbach2=gtodouble(gcbach2);
  if (!cmpis(D,-3) || !cmpis(D,-4))
  {
    p3=cgetg(5,t_VEC); p3[1]=un; p3[2]=lgetg(1,t_VEC);
    p3[3]=lgetg(1,t_VEC); p3[4]=un;
    return p3;
  }

  DATA = (qf_data *) xmalloc(sizeof(qf_data));
  DATA->D = D;
  DATA->precreg = 0;
  dr=cgetr(3); affir(D,dr); drc=rtodbl(dr); logd=log(fabs(drc));
  lim=sqrt(fabs(drc)/3.0); cp=(long)exp(sqrt(logd*log(logd)/8.0));

increaseimag:
  if (DEBUGLEVEL) { fprintferr("cbach = %f\n",cbach); flusherr(); }
  nreldep=nrelsup=0;
  limc=(long)(cbach*logd*logd);
  if (cp>limc) limc=cp;
  limc2=max(20,(long)(cbach2*logd*logd)); if (limc>limc2) limc2=limc;
  kc2=factorbasequad(D,limc2,limc,&numbase,&base,&kc,badprim,&nbbadprim);
  if (DEBUGLEVEL)
  { 
    fprintferr("temps creation de la factor base: ");
    fprintferr("%ld\n",timer2()); flusherr();
  }
  if (!kc)
  {
    free(base); free(numbase); avma=av; 
    cbach = check_bach(cbach,6.); cbach2=max(cbach2,cbach);
    goto increaseimag;
  }

  pro = (long*) xmalloc(sizeof(long)*(kc2+1));
  vectprime = (long*) xmalloc(sizeof(long)*(kc2+1));
  for (i=1; i<=kc2; i++)
  {
    p2=vectprime[i]=base[i];
    pro[i]=(p2>0) ? p2 : -p2;
  }
  free(base); base=pro;
  vperm = (long*) xmalloc(sizeof(long)*(kc+1));
  for (i=1; i<=kc; i++) vperm[i]=i;
  subbase=subfactorbase(&v,lim,kc,vectprime,vperm,&nbram,DATA);
  if (nbram<0)
  {
    free(vperm); free(vectprime); free(base); free(numbase); free(subbase);
    avma=av; cbach = check_bach(cbach,6.); cbach2=max(cbach2,cbach);
    goto increaseimag;
  }

  kcco=kc+CO;
  if (DEBUGLEVEL)
  { 
    fprintferr("KC = %ld, KCCO = %ld\n",kc,kcco);
    flusherr();
  }
  mglob=m=subbase[0];

  unform=cgetg(4,t_VEC);
  unform[1]=un; unform[3]=lshifti(absi(D),-2);
  unform[2]=mpodd(gmael(v,1,2)) ? un : zero;
  vp=powsubfact(v,m,CBUCH+7,unform,DATA);

  mat = (long **) xmalloc(sizeof(long*)*(kcco+1));
  matinit = (long **) xmalloc(sizeof(long*)*(kcco+1));
  for (i=1; i<=kcco; i++)
  {
    p1 = (long *) xmalloc(sizeof(long)*(kc+1));
    matinit[i]=mat[i]=p1;
    for (j=1; j<=kc; j++) p1[j]=0;
  }
  ex = (long*) xmalloc(sizeof(long)*(m+1));
  q=BITS_IN_RANDOM-1-(long)ceil(log((double)CBUCH)/log(2.0));
  s=0; nbtest=0;
  mm=m+nbram+CO;
  limhash=(limc<(MAXHALFULONG>>1))?limc*limc:((ulong)HIGHBIT)>>1;
  for (i=0; i<HASHT; i++) hashtab[i]=(long*)0;
  auxrel=0;
  while (s<mm)
  {
    for (i=1; i<=m; i++) ex[i]=mymyrand()>>q;
    av3=avma; form=vp[1][ex[1]];
    for (i=2; i<=m; i++) form=compimag(form,vp[i][ex[i]]);
    pc=primeform(D,stoi(base[1+(s%kc)]),0);
    form=compimag(form,pc);
    fpc=factorisequad(form,kc,limc,&lo,primfact,expoprimfact,badprim,nbbadprim,base,limhash);
    nbtest++;
    if (DEBUGLEVEL>=2 && !fpc) { fprintferr("."); flusherr(); }
    if (fpc>1)
    {
      fpd=largeprime(fpc,ex,1+(s%kc),0,mglob,hashtab);
      if (fpd)
      {
	auxrel++;
	s++;
	if (DEBUGLEVEL>=2) { fprintferr(" %ld",s); flusherr(); }
	lo1=lo;
	for (j=1; j<=lo1; j++)
	{ 
	  primfact1[j]=primfact[j];
	  expoprimfact1[j]=expoprimfact[j];
	}
	form1=vp[1][fpd[2]];
	for (i=2; i<=m; i++)
	  form1=compimag(form1,vp[i][fpd[i+1]]);
	if (fpd[m+2])
	{
	  pc=primeform(D,stoi(base[fpd[m+2]]),0);
	  form1=compimag(form1,pc);
	}
	b1=smodis((GEN)form1[2], fpc<<1);
	b2=smodis((GEN)form[2], fpc<<1);
	factorisequad(form1,kc,limc,&lo,primfact,expoprimfact,badprim,nbbadprim,base,limhash);
	if (b1==b2)
	{
	  for (i=1; i<=m; i++)
	    mat[s][numbase[subbase[i]]]=fpd[i+1]-ex[i];
	  mat[s][1+(s-1)%kc]--;
	  if (fpd[m+2]) mat[s][fpd[m+2]]++;
	  for (j=1; j<=lo; j++)
	  {
	    pp=primfact[j]; ep=expoprimfact[j];
	    b1=smodis((GEN)form1[2], pp<<1);
	    if (b1>pp) ep= -ep;
	    mat[s][numbase[pp]]-=ep;
	  }
	  for (j=1; j<=lo1; j++)
	  {
	    pp=primfact1[j]; ep=expoprimfact1[j];
	    b1=smodis((GEN)form[2], pp<<1);
	    if (b1>pp) ep= -ep;
	    mat[s][numbase[pp]]+=ep;
	  }
	}
	else
	{
	  if ((b1+b2)!=(fpc<<1)) { s--; auxrel--; }
	  else
	  {
	    for (i=1; i<=m; i++)
	      mat[s][numbase[subbase[i]]]= -fpd[i+1]-ex[i];
	    mat[s][1+(s-1)%kc]--;
	    if (fpd[m+2]) mat[s][fpd[m+2]]--;
	    for (j=1; j<=lo; j++)
	    {
	      pp=primfact[j]; ep=expoprimfact[j];
	      b1=smodis((GEN)form1[2], pp<<1);
	      if (b1>pp) ep= -ep;
	      mat[s][numbase[pp]]+=ep;
	    }
	    for (j=1; j<=lo1; j++)
	    {
	      pp=primfact1[j]; ep=expoprimfact1[j];
	      b1=smodis((GEN)form[2], pp<<1);
	      if (b1>pp) ep= -ep;
	      mat[s][numbase[pp]]+=ep;
	    }
	  }
	}
      }
    }	
    if (fpc==1)
    {
      s++;
      if (DEBUGLEVEL>=2) { fprintferr(" %ld",s); flusherr(); }
      for (i=1; i<=m; i++) mat[s][numbase[subbase[i]]]= -ex[i];
      mat[s][1+(s-1)%kc]--;
      for (j=1; j<=lo; j++)
      {
	pp=primfact[j]; ep=expoprimfact[j];
	b=smodis((GEN)form[2], pp<<1);
	if (b>pp) ep= -ep;
	mat[s][numbase[pp]]+=ep;
      }
    }
    avma=av3;
  }
  if (DEBUGLEVEL)
  {
    if (DEBUGLEVEL>=2) { fprintferr("\n"); flusherr(); }
    fprintferr("temps relations mm: %ld\n",timer2()); flusherr();
    fprintferr("Apres relations mm, s = nbtest = %ld\n",s); flusherr();
  }
  while (s<kcco)
  {
    for (i=1; i<=m; i++) ex[i]=mymyrand()>>q;
    av3=avma;
    form=vp[1][ex[1]];
    for (i=2; i<=m; i++) form=compimag(form,vp[i][ex[i]]);
    pc=primeform(D,stoi(base[s+1-CO]),0);
    form=compimag(form,pc);
    fpc=factorisequad(form,kc,limc,&lo,primfact,expoprimfact,badprim,nbbadprim,base,limhash);
    nbtest++;
    if (DEBUGLEVEL>=2 && !fpc) { fprintferr("."); flusherr(); }
    if (fpc>1)
    {
      fpd=largeprime(fpc,ex,s+1-CO,0,mglob,hashtab);
      if (fpd)
      {
	auxrel++;
	s++; lo1=lo;
	if (DEBUGLEVEL>=2) { fprintferr(" %ld",s); flusherr(); }
	for (j=1; j<=lo1; j++)
	{ 
	  primfact1[j]=primfact[j];
	  expoprimfact1[j]=expoprimfact[j];
	}
	form1=vp[1][fpd[2]];
	for (i=2; i<=m; i++)
	  form1=compimag(form1,vp[i][fpd[i+1]]);
	if (fpd[m+2])
	{
	  pc=primeform(D,stoi(base[fpd[m+2]]),0);
	  form1=compimag(form1,pc);
	}
	b1=smodis((GEN)form1[2], fpc<<1);
	b2=smodis((GEN)form[2], fpc<<1);
	factorisequad(form1,kc,limc,&lo,primfact,expoprimfact,badprim,nbbadprim,base,limhash);
	if (b1==b2)
	{
	  for (i=1; i<=m; i++)
	    mat[s][numbase[subbase[i]]]=fpd[i+1]-ex[i];
	  mat[s][s-CO]= -1;
	  if (fpd[m+2]) mat[s][fpd[m+2]]++;
	  for (j=1; j<=lo; j++)
	  {
	    pp=primfact[j]; ep=expoprimfact[j];
	    b1=smodis((GEN)form1[2], pp<<1);
	    if (b1>pp) ep= -ep;
	    mat[s][numbase[pp]]-=ep;
	  }
	  for (j=1; j<=lo1; j++)
	  {
	    pp=primfact1[j]; ep=expoprimfact1[j];
	    b1=smodis((GEN)form[2], pp<<1);
	    if (b1>pp) ep= -ep;
	    mat[s][numbase[pp]]+=ep;
	  }
	}
	else
	{
	  if ((b1+b2)!=(fpc<<1)) { s--; auxrel--; }
	  else
	  {
	    for (i=1; i<=m; i++)
	      mat[s][numbase[subbase[i]]]= -fpd[i+1]-ex[i];
	    mat[s][s-CO]= -1;
	    if (fpd[m+2]) mat[s][fpd[m+2]]--;
	    for (j=1; j<=lo; j++)
	    {
	      pp=primfact[j]; ep=expoprimfact[j];
	      b1=smodis((GEN)form1[2], pp<<1);
	      if (b1>pp) ep= -ep;
	      mat[s][numbase[pp]]+=ep;
	    }
	    for (j=1; j<=lo1; j++)
	    {
	      pp=primfact1[j]; ep=expoprimfact1[j];
	      b1=smodis((GEN)form[2], pp<<1);
	      if (b1>pp) ep= -ep;
	      mat[s][numbase[pp]]+=ep;
	    }
	  }
	}
      }
    }
    if (fpc==1)
    {
      s++;
      if (DEBUGLEVEL>=2) { fprintferr(" %ld",s); flusherr(); }
      for (i=1; i<=m; i++) mat[s][numbase[subbase[i]]]= -ex[i];
      mat[s][s-CO]= -1;
      for (j=1; j<=lo; j++)
      {
	pp=primfact[j]; ep=expoprimfact[j];
	b=smodis((GEN)form[2], pp<<1);
	if (b>pp) ep= -ep;
	mat[s][numbase[pp]]+=ep;
      }
      if (!(mat[s][s-CO])) { for (i=1; i<=kc; i++) mat[s][i]=0; s--; }
    }
    avma=av3;
  }
  if (DEBUGLEVEL)
  {
    if (DEBUGLEVEL>1) { fprintferr("\n"); flusherr(); }
    fprintferr("temps pour trouver les relations aleatoires: ");
    fprintferr("%ld\n",timer2()); flusherr();
    fprintferr("nbrelations/nbtest = %ld/%ld\n",s,nbtest); flusherr();
  }
  nbtest=auxrel=0; s=kc;
  if (DEBUGLEVEL>2)
  {
    if (kc2>kc)
    {
      fprintferr("be honest for primes in [%ld,%ld]\n",base[kc+1],base[kc2]);
      flusherr();
    }
  }
  while (s<kc2)
  {
    for (i=1; i<=m; i++) ex[i]=mymyrand()>>q;
    av3=avma;
    form=vp[1][ex[1]];
    for (i=2; i<=m; i++) form=compimag(form,vp[i][ex[i]]);
    pp=base[s+1];
    if (DEBUGLEVEL>=2) { fprintferr(" %ld",pp); flusherr(); }
    pc=primeform(D,stoi(pp),0);
    form=compimag(form,pc);
    fpc=factorisequad(form,s,pp-1,&lo,primfact,expoprimfact,badprim,nbbadprim,base,limhash);
    nbtest++; auxrel++;
    if (fpc==1) { s++; auxrel=0; }
    avma=av3;
    if (auxrel>20)
    {
      free(vperm); free(vectprime); free(ex); free(base);
      free(numbase); free(subbase);
      for (i=1; i<=kcco; i++) free(matinit[i]); free(matinit); free(mat);
      for (i=1; i<=m; i++) free(vp[i]); free(vp);
      for (i=1; i<HASHT; i++) freehash(hashtab[i]);
      avma=av; cbach = check_bach(cbach,6.); cbach2=max(cbach2,cbach);
      if (DEBUGLEVEL>=2) { fprintferr("\n"); flusherr(); }
      goto increaseimag;
    }
  }

  if (DEBUGLEVEL)
  {
    if (DEBUGLEVEL>=2) { fprintferr("\n"); flusherr(); }
    fprintferr("temps be honest: "); fprintferr("%ld\n",timer2()); flusherr();
  }
  matc=cgetg(kcco+1,t_MAT); for (i=1; i<=kcco; i++) matc[i]=lgetg(1,t_COL);
  mit=hnfspec(mat,&pdep,&matc,vperm,&matalpha,kcco,kc,m,&nlze,&col);
  if (DEBUGLEVEL)
  {
    fprintferr("temps hnfspec: "); fprintferr("%ld\n",timer2()); flusherr();
  }
  kccopro=kcco;
morerelimag:
  if (nlze)
  {
    long nlzebis;
    
    vinvperm=(long*)xmalloc(sizeof(long)*(kc+1));
    for (i=1; i<=kc; i++) vinvperm[vperm[i]]=i;
    s=0; extrarel=nlze+2;
    if (DEBUGLEVEL>1)
    {
      fprintferr("recherche de %ld relations supplementaires\n",extrarel);
      flusherr();
    }
    extramat=cgetg(extrarel+1,t_MAT);
    for (j=1; j<=extrarel; j++) extramat[j]=lgetg(kc+1,t_COL);
    nlzebis=min(nlze+1,kc);
    free(ex); ex=(long*)xmalloc(sizeof(long)*(nlzebis+1));
    while (s<extrarel)
    {
      for (i=1; i<=nlzebis; i++) ex[i]=mymyrand()>>q;
      form=gpuigs(primeform(D,stoi(labs(vectprime[vperm[1]])),0),ex[1]);
      for (i=2; i<=nlzebis; i++)
	form=compimag(form,gpuigs(primeform(D,stoi(labs(vectprime[vperm[i]])),0),ex[i]));
      fpc=factorisequad(form,kc,limc,&lo,primfact,expoprimfact,badprim,nbbadprim,base,limhash);
      if (DEBUGLEVEL>=2 && !fpc) { fprintferr("."); flusherr(); }
      if (fpc==1)
      {
	s++;
	if (DEBUGLEVEL>1) { fprintferr(" %ld",s); flusherr(); }
	p1=(GEN)extramat[s];
	for (i=1; i<=nlzebis; i++) p1[i]=lstoi(-ex[i]);
	for (i=nlzebis+1; i<=kc; i++) p1[i]=zero;
	for (j=1; j<=lo; j++)
	{
	  pp=primfact[j]; ep=expoprimfact[j];
	  b=smodis((GEN)form[2], pp<<1);
	  if (b>pp) ep= -ep;
	  k=vinvperm[numbase[pp]];
	  p1[k]=laddsg(ep,(GEN)p1[k]);
	}
	if (gcmp0(p1)) s--;
      }
    }
    extramatc=cgetg(extrarel+1,t_MAT);
    for (i=1; i<=extrarel; i++) extramatc[i]=lgetg(1,t_COL);
    if (DEBUGLEVEL)
    {
      if (DEBUGLEVEL>=2) { fprintferr("\n"); flusherr(); }
      fprintferr("temps calcul relations supplementaires: ");
      fprintferr("%ld\n",timer2()); flusherr();
    }
    if (nrelsup) nlze=0;
    mit=hnfadd(mit,&pdep,&matc,vperm,&matalpha,kccopro,kc,col,&nlze,extramat,extramatc);
    if (DEBUGLEVEL)
    { 
      fprintferr("temps hnfadd: ");
      fprintferr("%ld\n",timer2()); flusherr();
    }
    free(vinvperm); kccopro+=extrarel; col=kccopro-(lg(matalpha)-1);
    if (nlze)
    {
      nreldep++;
      if (nreldep>5)
      {
	free(vperm); free(vectprime); free(ex);
	free(base); free(numbase); free(subbase);
	for (i=1; i<=kcco; i++) free(matinit[i]);
	free(matinit); free(mat);
	for (i=1; i<=m; i++) free(vp[i]); free(vp);
	for (i=1; i<HASHT; i++) freehash(hashtab[i]);
	avma=av; cbach = check_bach(cbach,6.); cbach2=max(cbach2,cbach);
	goto increaseimag;
      }
      goto morerelimag;
    }
  }
  p1=gun; sizeofmit=lg(mit)-1;
  for (i=1; i<=sizeofmit; i++) p1=mulii(p1,gcoeff(mit,i,i));
  c_1=gdiv(gmul(p1,mppi(DEFAULTPREC)),
           gmul(lfunc(D),gsqrt(absi(D),DEFAULTPREC)));
  if (gcmpgs(gmul2n(c_1,2),3)<0) { c_1=stoi(10); nrelsup=1000; }
  if (gcmpgs(gmul2n(c_1,1),3)>0)
  {
    nrelsup++;
    if (nrelsup>7)
    {
      free(vperm); free(vectprime); free(ex);
      free(base); free(numbase); free(subbase);
      for (i=1; i<=kcco; i++) free(matinit[i]);
      free(matinit); free(mat);
      for (i=1; i<=m; i++) free(vp[i]); free(vp);
      for (i=1; i<HASHT; i++) freehash(hashtab[i]);
      if (cbach>5.99)
	err(warner,"check is greater than 1.5, suggest increasing extra relations");
      else
      {
	avma=av; cbach=min(2*cbach,6); cbach2=max(cbach2,cbach);
	goto increaseimag;
      }
    }
    else
    {
      if (DEBUGLEVEL)
      { 
	fprintferr("\n ***** check = %f\n\n",gtodouble(c_1));
	flusherr();
      }
      nlze=min(kc,nrelsup); goto morerelimag;
    }
  }
  u1u2=smith2(mit); u1=(GEN)u1u2[1]; met=(GEN)u1u2[3];
  u1=reducemodmatrix(ginv(u1),mit);
  lo=lg(met)-1;
  c=0; for (i=1; i<=lo; i++) if (!gcmp1(gcoeff(met,i,i))) c++;
  free(DATA);
  if (DEBUGLEVEL)
  { 
    fprintferr("temps smith/groupe de classes: ");
    fprintferr("%ld\n",timer2()); flusherr();
  }
  basecl=cgetg(c+1,t_VEC);
  for (j=1; j<=c; j++)
  {
    p3=gpui(primeform(D,stoi(labs(vectprime[vperm[1]])),0),gcoeff(u1,1,j),0);
    for (i=2; i<=lo; i++)
      p3=gmul(p3,gpui(primeform(D,stoi(labs(vectprime[vperm[i]])),0),
                      gcoeff(u1,i,j),0));
    basecl[j]=(long)p3;
  }
  if (DEBUGLEVEL)
  { 
    fprintferr("temps generateurs du groupe de classes: ");
    fprintferr("%ld\n",timer2()); flusherr();
  }
  tetpil=avma; p4=cgetg(5,t_VEC); p4[1]=lcopy(p1);
  mot=cgetg(c+1,t_VEC); p4[2]=(long)mot;
  for (i=1; i<=c; i++) mot[i]=lcopy(gcoeff(met,i,i));
  p4[3]=lcopy(basecl); p4[4]=lcopy(c_1);
  free(base); free(vectprime); free(ex);
  free(subbase); free(numbase); free(vperm);
  for (i=1; i<=kcco; i++) free(matinit[i]);
  free(matinit); free(mat);
  for (i=1; i<=m; i++) free(vp[i]); free(vp);
  for (i=1; i<HASHT; i++) freehash(hashtab[i]);
  if (DEBUGLEVEL)
  { 
    fprintferr("temps free finaux: ");
    fprintferr("%ld\n",timer2()); flusherr();
  }
  return gerepile(av,tetpil,p4);
}

static long
factorbasequad(GEN d, long n2, long n, long **ptnum, long **ptbase,
               long *ptkc, long *badprim, long *nbbadprim)
{
  byteptr delta=diffptr;
  long av2,i,pp,fl,kr,r,*numbase,*base;
  GEN p1;

  *ptnum = numbase = (long*) xmalloc(sizeof(long)*(n2+1));
  *ptbase =  base  = (long*) xmalloc(sizeof(long)*(n2+1));
  *ptkc=0; *nbbadprim=0; av2=avma; i=0; pp=*delta++; fl=1;
  while (pp<=n2)
  {
    kr=krogs(d,pp);
    if (kr!=-1)
    {
      if (kr) { i++; numbase[pp]=i; base[i]=pp; }
      else
      {
	p1=divis(d,pp);
	if (smodis(p1,pp)) { i++; numbase[pp]=i; base[i]= -pp; }
	else
	{
	  if (pp==2)
	  {
	    r=mod8(p1); if (signe(d)<0 && r) r=8-r;
	    if (r>=4) { i++; numbase[pp]=i; base[i]= -pp; }
	    else badprim[++(*nbbadprim)]=pp;
	  }
	  else badprim[++(*nbbadprim)]=pp;
	}
      }
    }
    pp+=*delta++;
    if (pp>n && fl) { *ptkc=i; fl=0; }
  }
  avma=av2; return i;
}

static long
factorisequad(GEN f, long n, long limp, long *ptlo, long *primfact,
              long *expoprimfact, long *badprim, long nbbadprim, long *base,
	      long limhash)
{
  long i,p,k,av1,av2,lo;
  GEN x,q,r;

  av1=avma; lo=0; x=absi((GEN)f[1]);
  if (gcmp1(x)) { avma=av1; *ptlo=0; return 1; }

  av2=avma;
  for (i=1; i<=n; i++)
  {
    p=base[i]; q=dvmdis(x,p,&r);
    if (!signe(r))
    {
      primfact[++lo]=p; x=q; k=0; av2=avma;
      for(;;)
      { 
	k++; q=dvmdis(x,p,&r);
	if (signe(r)) break;
	x=q;
      }
      av2=avma; expoprimfact[lo]=k;
    }
    else avma=av2;
    if (cmpis(q,p)<=0)
    {
      if (gcmp1(x)) { avma=av1; *ptlo=lo; return 1; }
      if (cmpis(x,limp)<=0)
      {
	for (i=1; i<=nbbadprim; i++)
	  if (!smodis(x,badprim[i])) { avma=av1; *ptlo=lo; return 0; }
	primfact[++lo]=itos(x); expoprimfact[lo]=1;
	avma=av1; *ptlo=lo; return 1;
      }
    }
  }
  *ptlo=lo; avma=av1; /* valid since cmpis and itos don't touch the stack */
  if (cmpis(x,limhash)<=0) return itos(x);
  return 0;
}

GEN
buchreal(GEN D, GEN gsens, GEN gcbach, GEN gcbach2, GEN gRELSUP, long prec)
{
  long sens,RELSUP,nrelsup,nreldep,kccopro;
  double cbach,cbach2;
  long precreg,limc,limbach,cardsubbase,cardsub,nbram,auxrel,e,maxe;
  long nbprimfact,nbprimfact1,lo,initform5,findecycle;
  long nbrhocourant,nbrhocumule,nrho,*fpd,b1,b2,fpc,pp,ep,b,extrarel,c,limhash;
  long sizebach,*numbase,*base,*subbase;
  long **mat,**matinit,*vectprime,*vperm,*vinvperm;
  long primfact[100],expoprimfact[100],primfact1[100],expoprimfact1[100];
  long badprim[100],nbbadprim,nlze,nlzecard,col,sizeofmit;
  long av=avma,av1,av2,tetpil,sizemat,sizematcol,i,j,k,*pro,p2,*p1,*ex,qqq,s,nbtest,mm;
  double drc,lim,logd;
  GEN dr,sqrtD,isqrtD,log2precis,tabprform,unform;
  GEN matc,extramat,extramatc,**tabpowprform;
  GEN forminit,form,form0,form1,form2,pc,mit,met,mot,p3,p4,pdep,matalpha;
  GEN vecexpo,reg,c_1,u1u2,u1,basecl;
  long* hashtab[HASHT];
  qf_data *DATA;

  if (DEBUGLEVEL) timer2();
  if (typ(D)!=t_INT || typ(gsens)!=t_INT || typ(gRELSUP)!=t_INT)
    err(typeer,"buchreal");
  if (!signe(D)) err(bucher2);
  if (signe(D)<0) 
    err(talker,"negative discriminant in buchreal, try buchimag");
  s=mod4(D); if (s==2 || s==3) err(funder2,"buchreal");
  sens = signe(gsens) ? 1 : 0;
  RELSUP=itos(gRELSUP);
  cbach=gtodouble(gcbach); cbach2=gtodouble(gcbach2);

                   /* Initialisation : */

 /* precision en digits decimaux=2*(#digits decimaux de D)+50
  * ici CBUCH=15,q=27
  * on prendra les p decomposes tels que prod(p)>lim dans la subbase
  * limc=Max(c.(log(D))^2,exp((1/8).sqrt(log(D).loglog(D))))
  * limbach=Max(6.(log(D))^2,exp((1/8).sqrt(log(D).loglog(D))))
  * subbase contient les p decomposes tels que prod(p)>sqrt(D)
  * cardsubbase=cardsub=subbase[0]=#subbase;
  * tabprform est la table des form[p] pour p dans subbase
  * nbram est le nombre de p divisant D elimines dans subbase
  * tabpowprform est la table des puissances des formes dans tabprform
  */

  qqq=BITS_IN_RANDOM-1-(long)ceil(log((double)CBUCH)/log(2.0));
  dr=cgetr(3); affir(D,dr);
  drc=rtodbl(dr); lim=sqrt(drc); logd=log(drc);

  DATA = (qf_data *) xmalloc(sizeof(qf_data));
  DATA->D=D;
  DATA->sens=sens;

increasereal:
  if (DEBUGLEVEL) { fprintferr("cbach = %f\n",cbach); flusherr(); }
  nreldep=nrelsup=0;

  precreg=max(prec+1,2*(gexpo(D)>>TWOPOTBITS_IN_LONG)+5);
  sqrtD=gsqrt(D,precreg); isqrtD=gfloor(sqrtD);
  log2precis=glog(gdeux,precreg);
  limc=max((long)(cbach*logd*logd),13);
  i=(long)exp(sqrt(logd*log(logd)/8.0));
  if (i> limc) limc=i;
  limbach = max(20,(long)(cbach2*logd*logd));
  if (limc > limbach) limbach=limc;
  DATA->isqrtD=isqrtD;
  DATA->sqrtD=sqrtD;
  DATA->precreg=precreg;

  sizebach=factorbasequad(D,limbach,limc,&numbase,&base,&sizemat,badprim,&nbbadprim);
  if (DEBUGLEVEL)
  { 
    fprintferr("temps creation de la factor base: ");
    fprintferr("%ld\n",timer2()); flusherr();
  }
  if (!sizemat)
  {
    free(base); free(numbase);
    avma=av; cbach = check_bach(cbach,6.); cbach2=max(cbach2,cbach);
    goto increasereal;
  }
  if (DEBUGLEVEL>7)
  {
    fprintferr("factor base :\n");
    for (i=1; i<=sizemat; i++) fprintferr("%ld ",base[i]);
    fprintferr("\n"); flusherr();
  }
  pro = (long*) xmalloc(sizeof(long)*(sizebach+1));
  vectprime = (long*) xmalloc(sizeof(long)*(sizebach+1));
  for (i=1; i<=sizebach; i++)
  { 
    p2=vectprime[i]=base[i];
    pro[i]=(p2>0) ? p2 : -p2;
  }
  free(base); base=pro;
  vperm = (long*) xmalloc(sizeof(long)*(sizemat+1));
  for (i=1; i<=sizemat; i++) vperm[i]=i;
  subbase=subfactorbase(&tabprform,lim,sizemat,vectprime,vperm,&nbram,DATA);
  if (nbram<0)
  {
    free(vperm); free(vectprime); free(base); free(numbase); free(subbase);
    avma=av; cbach = check_bach(cbach,6.); cbach2=max(cbach2,cbach);
    goto increasereal;
  }
  cardsubbase=cardsub=subbase[0];
  if (DEBUGLEVEL>7)
  {
    fprintferr("sub factor base : ");
    for (i=1; i<=cardsubbase; i++)
    {
      fprintferr("%ld : ",subbase[i]); outerr((GEN)tabprform[i]);
    }
    fprintferr("\n"); flusherr();
  }
  sizematcol=sizemat+RELSUP;
  if (DEBUGLEVEL)
  { 
    fprintferr("KC = %ld, KCCO = %ld\n",sizemat,sizematcol);
    flusherr();
  }

  unform=cgetg(6,t_VEC);
  unform[1]=un;
  if (mpodd(gmael(tabprform,1,2)))
    unform[2]= (mpodd(isqrtD)) ? (long) isqrtD : laddsi(-1,isqrtD);
  else
  { 
    unform[2] = (long) isqrtD; 
    (((GEN)unform[2])[lgefint(isqrtD)-1]) &= (~0x1);
  }
  unform[3]=lshifti(subii(sqri((GEN)unform[2]),D),-2);
  unform[4]=zero; unform[5]=lgetr(precreg); affsr(1,(GEN)unform[5]);
  tabpowprform=powsubfact(tabprform,cardsub,CBUCH+7,unform,DATA);

  vecexpo=cgetg(sizematcol+1,t_VEC);
  for (i=1; i<=sizematcol; i++) vecexpo[i]=lgetr(precreg);
  mat = (long**) xmalloc((sizematcol+1)*sizeof(long*));
  matinit = (long**) xmalloc((sizematcol+1)*sizeof(long*));
  for (i=1; i<=sizematcol; i++)
    matinit[i]=mat[i] = (long*) xmalloc((sizemat+1)*sizeof(long));

  for (i=1; i<=sizematcol; i++)
    for (j=1; j<=sizemat; j++) mat[i][j]=0;
  ex = (long*) xmalloc(sizeof(long)*(cardsub+1));
  limhash=(limc<(((ulong)MAXHALFULONG)>>1))?limc*limc:((ulong)HIGHBIT)>>1;
  s=nbtest=auxrel=0;
  mm=cardsub+nbram+RELSUP;
  for (i=0; i<HASHT; i++) hashtab[i]=NULL;

  while (s<mm)
  {
    for (i=1; i<=cardsub; i++) ex[i]=mymyrand()>>qqq;
    av1=avma;
    form = form0 = initializeform3(ex,tabpowprform,cardsub,DATA);
    initform5 = findecycle = nbrhocourant = nbrhocumule = 0;
    while (s<mm)
    {
      GEN *gptr[2];

      av2=avma;
      fpc=factorisequad(form,sizemat,limc,&nbprimfact,primfact,expoprimfact,badprim,nbbadprim,base,limhash);
      if (DEBUGLEVEL>=2 && !fpc) { fprintferr("."); flusherr(); }
      if (fpc==1)
      {
	if (!initform5)
	{
	  form1=initializeform5(ex,tabpowprform,cardsub,DATA);
	  initform5=1;
	}
	for (i=1; i<=nbrhocourant; i++)
	  form1=rhorealform5(form1,DATA);
	nbrhocourant=0; s++;
	if (DEBUGLEVEL>=2) { fprintferr(" %ld",s); flusherr(); }

                          /* addcolumnmat */

	for (i=1; i<=cardsub; i++) mat[s][numbase[subbase[i]]]= -ex[i];
	for (j=1; j<=nbprimfact; j++)
	{
	  pp=primfact[j]; ep=expoprimfact[j];
	  b=smodis((GEN)form1[2], pp<<1);
	  if (b>pp) ep= -ep; mat[s][numbase[pp]]+=ep;
	}
	affrr(shiftr(mpadd(mulir(mulsi(EXP220,(GEN)form1[4]),log2precis),
			   mplog(absr((GEN)form1[5]))),-1),(GEN)vecexpo[s]);
      }
      else if (fpc>1)
      {
	fpd=largeprime(fpc,ex,0,nbrhocumule,cardsubbase,hashtab);
	if (fpd)
	{
	  if (!initform5)
	  {
	    form1=initializeform5(ex,tabpowprform,cardsub,DATA);
	    initform5=1;
	  }
	  for (i=1; i<=nbrhocourant; i++)
	    form1=rhorealform5(form1,DATA);
	  nbrhocourant=0; nbprimfact1=nbprimfact;
	  for (j=1; j<=nbprimfact1; j++)
	  { 
	    primfact1[j]=primfact[j]; 
	    expoprimfact1[j]=expoprimfact[j];
	  }
	  form2=tabpowprform[1][fpd[2]];
	  for (i=2; i<=cardsub; i++)
	    form2=comprealform5(form2,tabpowprform[i][fpd[i+1]],DATA);
	  for (i=1; i<=fpd[cardsub+3]; i++)
	    form2=rhorealform5(form2,DATA);
	  if (!sens && signe(addii((GEN)form2[1],(GEN)form2[3])))
	  { 
	    setsigne(form2[1],1);
	    setsigne(form2[3],-1);
	  }
	  b1=smodis((GEN)form2[2], fpc<<1);
	  b2=smodis((GEN)form1[2], fpc<<1);
	  factorisequad(form2,sizemat,limc,&nbprimfact,primfact,expoprimfact,badprim,nbbadprim,base,limhash);
	  if (b1==b2)
	  {
	                 /* addcolumnmat1 */
	    s++;
	    if (DEBUGLEVEL>=2) fprintferr(" %ld",s);

	    for (i=1; i<=cardsub; i++)
	      mat[s][numbase[subbase[i]]]=fpd[i+1]-ex[i];
	    for (j=1; j<=nbprimfact; j++)
	    {
	      pp=primfact[j]; ep=expoprimfact[j];
	      b1=smodis((GEN)form2[2], pp<<1);
	      if (b1>pp) ep= -ep; mat[s][numbase[pp]]-=ep;
	    }
	    for (j=1; j<=nbprimfact1; j++)
	    {
	      pp=primfact1[j]; ep=expoprimfact1[j];
	      b1=smodis((GEN)form1[2], pp<<1);
	      if (b1>pp) ep= -ep; mat[s][numbase[pp]]+=ep;
	    }
	    affrr(shiftr(mpadd(mulir(mulsi(EXP220,subii((GEN)form1[4],(GEN)form2[4])),log2precis), mplog(absr(divrr((GEN)form1[5],(GEN)form2[5])))),-1), (GEN)vecexpo[s]);
	  }
	  else if ((b1+b2)==(fpc<<1))
	  {                
		       /* addcolumnmat2 */
	    s++;
	    if (DEBUGLEVEL>=2) { fprintferr(" %ld",s); flusherr(); }

	    for (i=1; i<=cardsub; i++)
	      mat[s][numbase[subbase[i]]]=-fpd[i+1]-ex[i];
	    for (j=1; j<=nbprimfact; j++)
	    {
	      pp=primfact[j]; ep=expoprimfact[j];
	      b1=smodis((GEN)form2[2], pp<<1);
	      if (b1>pp) ep= -ep; mat[s][numbase[pp]]+=ep;
	    }
	    for (j=1; j<=nbprimfact1; j++)
	    {
	      pp=primfact1[j]; ep=expoprimfact1[j];
	      b1=smodis((GEN)form1[2], pp<<1);
	      if (b1>pp) ep= -ep; mat[s][numbase[pp]]+=ep;
	    }
	    affrr(shiftr(mpadd(mulir(mulsi(EXP220,addii((GEN)form1[4],(GEN)form2[4])),log2precis),mplog(absr(mulrr((GEN)form1[5],(GEN)form2[5])))),-1),(GEN)vecexpo[s]);
	  }
	}
      }
      form=rhorealform3(form,DATA);
      nbrhocourant++; nbrhocumule++;

      if (findecycle) { avma=av2; break; }

      if (!sens && gegal((GEN)form[1],negi((GEN)form0[1])) && 
	  gegal((GEN)form[2],(GEN)form0[2])) findecycle=1;

      else if (gegal((GEN)form[1],(GEN)form0[1]) && 
	  gegal((GEN)form[2],(GEN)form0[2])) findecycle=1;

      tetpil=avma;
      gptr[0]=&form; gptr[1]=&form1;
      gerepilemany(av2,gptr,initform5?2:1);
    }
    avma=av1;
  }

  if (DEBUGLEVEL)
  {
    if (DEBUGLEVEL>=2) fprintferr("\n");
    fprintferr("temps relations mm: ");
    fprintferr("%ld\n",timer2());
    fprintferr("Apres relations mm, s = nbtest = %ld\n",s);
    flusherr();
  }

  while (s<sizematcol)
  {
    for (i=1; i<=cardsub; i++) ex[i]=mymyrand()>>qqq;
    av1=avma;
    form=initializeform3(ex,tabpowprform,cardsub,DATA);
    pc=redrealform3(primeform(D,stoi(base[s+1-RELSUP]),precreg),DATA);
    form=comprealform3(form,pc,DATA); forminit=form; nrho=0;

    for(;;)
    {
      fpc=factorisequad(form,sizemat,limc,&nbprimfact,primfact,expoprimfact,badprim,nbbadprim,base,limhash);
      nbtest++;
      if (fpc == 1)
      {
	form=initializeform5(ex,tabpowprform,cardsub,DATA);
	pc=redrealform(primeform(D,stoi(base[s+1-RELSUP]),precreg),DATA);
	pc[4]=zero; affsr(1,(GEN)pc[5]);
	form=comprealform5(form,pc,DATA);
	for (i=1; i<=nrho; i++) form=rhorealform5(form,DATA);
	s++;
	if (DEBUGLEVEL>=2) { fprintferr(" %ld",s); flusherr(); }
	for (i=1; i<=cardsub; i++) mat[s][numbase[subbase[i]]]= -ex[i];
	mat[s][s-RELSUP]= -1;
	for (j=1; j<=nbprimfact; j++)
	{
	  pp=primfact[j]; ep=expoprimfact[j];
	  b=smodis((GEN)form[2], pp<<1);
	  if (b>pp) ep= -ep; mat[s][numbase[pp]]+=ep;
	}
	affrr(shiftr(mpadd(mulir(mulsi(EXP220,(GEN)form[4]),log2precis),mplog(absr((GEN)form[5]))),-1),(GEN)vecexpo[s]);
	if (!(mat[s][s-RELSUP]))
	{ 
	  for (i=1; i<=sizemat; i++) mat[s][i]=0;
	  s--;
	}

	break;
      }

      if (DEBUGLEVEL>=2 && !fpc) { fprintferr("."); flusherr(); }
      if (fpc)
      {
	fpd=largeprime(fpc,ex,s+1-RELSUP,nrho,cardsubbase,hashtab);
	if (fpd)
	{
	  form=initializeform5(ex,tabpowprform,cardsub,DATA);
	  pc=redrealform(primeform(D,stoi(base[s+1-RELSUP]),precreg),DATA);
	  pc[4]=zero; affsr(1,(GEN)pc[5]);
	  form=comprealform5(form,pc,DATA);
	  for (i=1; i<=nrho; i++) form=rhorealform5(form,DATA);
	  auxrel++; s++; nbprimfact1=nbprimfact;
	  if (DEBUGLEVEL>=2) { fprintferr(" %ld",s); flusherr(); }
	  for (j=1; j<=nbprimfact1; j++)
	  {
	    primfact1[j]=primfact[j];
	    expoprimfact1[j]=expoprimfact[j];
	  }
	  form1=tabpowprform[1][fpd[2]];
	  for (i=2; i<=cardsub; i++)
	    form1=comprealform5(form1,tabpowprform[i][fpd[i+1]],DATA);
	  if (fpd[cardsub+2])
	  {
	    pc=redrealform(primeform(D,stoi(base[fpd[cardsub+2]]),precreg),DATA);
	    pc[4]=zero; affsr(1,(GEN)pc[5]);
	    form1=comprealform5(form1,pc,DATA);
	  }
	  for (i=1; i<=fpd[cardsub+3]; i++)
	    form1=rhorealform5(form1,DATA);
	  if (!sens && signe(addii((GEN)form1[1],(GEN)form1[3])))
	  {
	    setsigne(form1[1],1);
	    setsigne(form1[3],-1);
	  }
	  b1=smodis((GEN)form1[2], fpc<<1);
	  b2=smodis((GEN)form[2], fpc<<1);
	  factorisequad(form1,sizemat,limc,&nbprimfact,primfact,expoprimfact,badprim,nbbadprim,base,limhash);
	  if (b1==b2)
	  {
	    for (i=1; i<=cardsub; i++)
	      mat[s][numbase[subbase[i]]] = fpd[i+1]-ex[i];
	    mat[s][s-RELSUP]= -1;
	    if (fpd[cardsub+2]) mat[s][fpd[cardsub+2]]++;
	    for (j=1; j<=nbprimfact; j++)
	    {
	      pp=primfact[j]; ep=expoprimfact[j];
	      b1=smodis((GEN)form1[2], pp<<1);
	      if (b1>pp) ep= -ep; mat[s][numbase[pp]]-=ep;
	    }
	    for (j=1; j<=nbprimfact1; j++)
	    {
	      pp=primfact1[j]; ep=expoprimfact1[j];
	      b1=smodis((GEN)form[2], pp<<1);
	      if (b1>pp) ep= -ep; mat[s][numbase[pp]]+=ep;
	    }
	    affrr(shiftr(mpadd(mulir(mulsi(EXP220,subii((GEN)form[4],(GEN)form1[4])),log2precis),mplog(absr(divrr((GEN)form[5],(GEN)form1[5])))),-1),(GEN)vecexpo[s]);
	    break;
	  }
	  else
	  {
	    if ((b1+b2)!=(fpc<<1)) { s--; auxrel--; }
	    else
	    {
	      for (i=1; i<=cardsub; i++)
		mat[s][numbase[subbase[i]]] = -fpd[i+1]-ex[i];
	      mat[s][s-RELSUP]= -1;
	      if (fpd[cardsub+2]) mat[s][fpd[cardsub+2]]--;
	      for (j=1; j<=nbprimfact; j++)
	      {
		pp=primfact[j]; ep=expoprimfact[j];
		b1=smodis((GEN)form1[2], pp<<1);
		if (b1>pp) ep= -ep; mat[s][numbase[pp]]+=ep;
	      }
	      for (j=1; j<=nbprimfact1; j++)
	      {
		pp=primfact1[j]; ep=expoprimfact1[j];
		b1=smodis((GEN)form[2], pp<<1);
		if (b1>pp) ep= -ep; mat[s][numbase[pp]]+=ep;
	      }
	      affrr(shiftr(mpadd(mulir(mulsi(EXP220,addii((GEN)form1[4],(GEN)form[4])),log2precis),mplog(absr(mulrr((GEN)form1[5],(GEN)form[5])))),-1),(GEN)vecexpo[s]);
	      break;
	    }
	  }
	}
      }
      form=rhorealform3(form,DATA); nrho++;
      if (sens || !signe(addii((GEN)form[1],(GEN)form[3])))
      { 
	form=rhorealform3(form,DATA); nrho++;
      }
      else 
      { 
	setsigne(form[1],1);
	setsigne(form[3],-1);
      }
      if ( gegal((GEN)form[1],(GEN)forminit[1])
	&& gegal((GEN)form[2],(GEN)forminit[2])
	&& gegal((GEN)form[3],(GEN)forminit[3])) break;
    }
    avma=av1;
  }

  if (DEBUGLEVEL)
  {
    if (DEBUGLEVEL>1) { fprintferr("\n"); flusherr(); }
    fprintferr("temps pour trouver les relations aleatoires: ");
    fprintferr("%ld\n",timer2()); flusherr();
    fprintferr("nbrelations/nbtest = %ld/%ld\n",s,nbtest); flusherr();
  }

  nbtest=auxrel=0; s=sizemat;
  if (DEBUGLEVEL>2)
  {
    if (sizebach>sizemat)
    {
      fprintferr("be honest for primes from %ld to %ld\n",
                  base[sizemat+1],base[sizebach]);
      flusherr();
    }
  }

  while (s<sizebach)
  {
    for (i=1; i<=cardsub; i++) ex[i]=mymyrand()>>qqq;
    av1=avma;
    form=initializeform3(ex,tabpowprform,cardsub,DATA);
    pp=base[s+1];
    if (DEBUGLEVEL>=2) { fprintferr(" %ld",pp); flusherr(); }
    pc=redrealform3(primeform(D,stoi(pp),precreg),DATA);
    form=comprealform3(form,pc,DATA); forminit=form;
    for(;;)
    {
      fpc=factorisequad(form,s,pp-1,&nbprimfact,primfact,expoprimfact,badprim,nbbadprim,base,limhash);
      nbtest++;
      if (fpc == 1)
      { 
	auxrel=0; s++; break;
      }

      form=rhorealform3(form,DATA);
      if (sens || !signe(addii((GEN)form[1],(GEN)form[3])))
	form=rhorealform3(form,DATA);
      else
      { 
	setsigne(form[1],1);
	setsigne(form[3],-1);
      }

      if (gegal((GEN)form[1],(GEN)forminit[1])
       && gegal((GEN)form[2],(GEN)forminit[2])
       && gegal((GEN)form[3],(GEN)forminit[3])) break;
    }

    if (auxrel>20)
    {
      free(vperm); free(vectprime); free(ex);
      free(base); free(numbase); free(subbase);
      for (i=1; i<=sizematcol; i++) free(matinit[i]);
      free(matinit); free(mat);
      for (i=1; i<=cardsub; i++) free(tabpowprform[i]);
      free(tabpowprform);
      for (i=1; i<HASHT; i++) freehash(hashtab[i]);
      avma=av; cbach = check_bach(cbach,6.); cbach2=max(cbach2,cbach);
      if (DEBUGLEVEL>=2) { fprintferr("\n"); flusherr(); }
      goto increasereal;
    }
    avma=av1;
  }

  if (DEBUGLEVEL)
  {
    if (DEBUGLEVEL>=2) fprintferr("\n");
    fprintferr("temps be honest: ");
    fprintferr("%ld\n",timer2()); flusherr();
  }
  matc=cgetg(sizematcol+1,t_MAT);
  for (i=1; i<=sizematcol; i++)
  { 
    p1=cgetg(2,t_COL); matc[i]=(long)p1; p1[1]=vecexpo[i];
  }
  if (DEBUGLEVEL>7)
  {
    fprintferr("matc : "); outerr(matc); flusherr();
  }
  mit=hnfspec(mat,&pdep,&matc,vperm,&matalpha,sizematcol,sizemat,cardsub,&nlze,&col);
  if (DEBUGLEVEL)
  {
    fprintferr("temps hnfspec: "); fprintferr("%ld\n",timer2()); flusherr();
  }
  if (DEBUGLEVEL>7)
  {
    fprintferr("matc : "); outerr(matc); flusherr();
  }
  kccopro=sizematcol;

morerelreal:
  if (nlze)
  {
    vinvperm = (long*) xmalloc(sizeof(long)*(sizemat+1));
    for (i=1; i<=sizemat; i++) vinvperm[vperm[i]]=i;
    s=0; extrarel=nlze+2;
    if (DEBUGLEVEL>1)
    {
      fprintferr("recherche de %ld relations supplementaires\n",extrarel);
      flusherr();
    }
    extramat=cgetg(extrarel+1,t_MAT);
    for (j=1; j<=extrarel; j++) extramat[j]=lgetg(sizemat+1,t_COL);
    extramatc=cgetg(extrarel+1,t_MAT);
    for (i=1; i<=extrarel; i++) extramatc[i]=lgetg(2,t_COL);
    free(ex); nlzecard=max(nlze,cardsub);
    ex = (long*) xmalloc(sizeof(long)*(nlzecard+1));
    while (s<extrarel)
    {
      for (i=1; i<=nlzecard; i++) ex[i]=mymyrand()>>qqq;
      form=gpuigs(primeform(D,stoi(labs(vectprime[vperm[1]])),precreg),ex[1]);
      for (i=2; i<=nlzecard; i++)
	form=compreal(form,gpuigs(primeform(D,stoi(labs(vectprime[vperm[i]])),precreg),ex[i]));
      fpc=factorisequad(form,sizemat,limc,&lo,primfact,expoprimfact,badprim,nbbadprim,base,limhash);
      if (DEBUGLEVEL>=2 && !fpc) { fprintferr("."); flusherr(); }

      if (fpc==1)
      {
	s++; p1=(GEN)extramat[s];
	if (DEBUGLEVEL>=2) { fprintferr(" %ld",s); flusherr(); }
	for (i=1; i<=nlzecard; i++) p1[i]=lstoi(-ex[i]);
	for (i=nlzecard+1; i<=sizemat; i++) p1[i]=zero;
	for (j=1; j<=lo; j++)
	{
	  pp=primfact[j]; ep=expoprimfact[j];
	  b=smodis((GEN)form[2], pp<<1);
	  if (b>pp) ep= -ep;
	  k=vinvperm[numbase[pp]];
	  p1[k]=laddsg(ep,(GEN)p1[k]);
	}
	if (gcmp0(p1)) s--;
	else coeff(extramatc,1,s)=form[4];
      }
    }

    if (DEBUGLEVEL)
    {
      if (DEBUGLEVEL>=2) fprintferr("\n");
      fprintferr("temps calcul relations supplementaires: ");
      fprintferr("%ld\n",timer2()); flusherr();
    }
    if (nrelsup) nlze=0;
    mit=hnfadd(mit,&pdep,&matc,vperm,&matalpha,kccopro,sizemat,col,&nlze,extramat,extramatc);
    if (DEBUGLEVEL)
    { 
      fprintferr("temps hnfadd: ");
      fprintferr("%ld\n",timer2()); flusherr();
    }
    if (DEBUGLEVEL>7)
    {
      fprintferr("mit : "); outerr(mit); flusherr();
      fprintferr("matc : "); outerr(matc); flusherr();
    }
    free(vinvperm); kccopro+=extrarel; col=kccopro-(lg(matalpha)-1);
    if (nlze)
    {
      nreldep++;
      if (nreldep>5)
      {
	free(vperm); free(vectprime); free(ex);
	free(base); free(numbase); free(subbase);
	for (i=1; i<=sizematcol; i++) free(matinit[i]);
	free(matinit); free(mat);
	for (i=1; i<=cardsub; i++) free(tabpowprform[i]);
	free(tabpowprform);
	for (i=1; i<HASHT; i++) freehash(hashtab[i]);
	avma=av; cbach = check_bach(cbach,6.); cbach2=max(cbach2,cbach);
	goto increasereal;
      }
      goto morerelreal;
    }
  }
  p1=gun; sizeofmit=lg(mit)-1;
  for (i=1; i<=sizeofmit; i++) p1=mulii(p1,gcoeff(mit,i,i));
  extrarel=col-sizeofmit;
  reg=gabs(gcoeff(matc,1,1),0); e=maxe=0;
  for (i=2; i<=extrarel && maxe<=0; i++)
  {
    reg=gcdrealnoer(gcoeff(matc,1,i),reg,&e);
    maxe=(maxe)? max(maxe,e): e;
  }
  if (DEBUGLEVEL>7) { fprintferr("reg = "); outerr(reg); flusherr(); }
  if (maxe>0)
  {
    /* c'est idiot de recalculer. A changer */
    prec=(precreg<<1)-3;
    free(vperm); free(vectprime); free(ex);
    free(base); free(numbase); free(subbase);
    for (i=1; i<=sizematcol; i++) free(matinit[i]);
    free(matinit); free(mat);
    for (i=1; i<=cardsub; i++) free(tabpowprform[i]);
    free(tabpowprform);
    for (i=1; i<HASHT; i++) freehash(hashtab[i]);
    avma=av; goto increasereal;
  }
  if (DEBUGLEVEL)
  { 
    fprintferr("temps regulateur: ");
    fprintferr("%ld\n",timer2()); flusherr();
  }
  if (gexpo(reg)<=-3)
  {
    nrelsup++;
    if (nrelsup>7)
    {
      free(vperm); free(vectprime); free(ex);
      free(base); free(numbase); free(subbase);
      for (i=1; i<=sizematcol; i++) free(matinit[i]);
      free(matinit); free(mat);
      for (i=1; i<=cardsub; i++) free(tabpowprform[i]);
      free(tabpowprform);
      for (i=1; i<HASHT; i++) freehash(hashtab[i]);
      avma=av; cbach = check_bach(cbach,6.); cbach2=max(cbach2,cbach);
      goto increasereal;
    }
    nlze=min(sizemat,nrelsup);
    if (DEBUGLEVEL) { fprintferr("regulateur nul : \n"); flusherr(); }
    goto morerelreal;
  }

  c_1=gdiv(gmul2n(gmul(p1,reg),1),gmul(lfunc(D),sqrtD));
  if (gcmpgs(gmul2n(c_1,2),3)<0) { c_1=stoi(10); nrelsup=1000; }
  if (gcmpgs(gmul2n(c_1,1),3)>0)
  {
    nrelsup++;
    if (nrelsup>7)
    {
      free(vperm); free(vectprime); free(ex);
      free(base); free(numbase); free(subbase);
      for (i=1; i<=sizematcol; i++) free(matinit[i]);
      free(matinit); free(mat);
      for (i=1; i<=cardsub; i++) free(tabpowprform[i]);
      free(tabpowprform);
      for (i=1; i<HASHT; i++) freehash(hashtab[i]);
      if (cbach>5.99)
	err(warner,"check is greater than 1.5, suggest increasing extra relations");
      else
      {
	avma=av; cbach=min(2*cbach,6); cbach2=max(cbach2,cbach);
	goto increasereal;
      }
    }
    else
    {
      nlze=min(sizemat,nrelsup);
      if (DEBUGLEVEL)
      { 
	fprintferr("\n ***** check = %f\n\n",gtodouble(c_1));
	flusherr();
      }
      goto morerelreal;
    }
  }
  u1u2=smith2(mit); u1=(GEN)u1u2[1];
  met=(GEN)u1u2[3]; u1=reducemodmatrix(ginv(u1),mit);
  lo=lg(met)-1; c=0;
    for (i=1; i<=lo; i++)
      if (!gcmp1(gcoeff(met,i,i))) c++;
  free(DATA);
  if (DEBUGLEVEL)
  { 
    fprintferr("temps smith/groupe de classes: ");
    fprintferr("%ld\n",timer2()); flusherr();
  }
  basecl=cgetg(c+1,t_VEC);
  for (j=1; j<=c; j++)
  {
    p3=gpui(primeform(D,stoi(labs(vectprime[vperm[1]])),precreg),gcoeff(u1,1,j),precreg);
    for (i=2; i<=lo; i++)
      p3=gmul(p3,gpui(primeform(D,stoi(labs(vectprime[vperm[i]])),precreg),gcoeff(u1,i,j),precreg));
    basecl[j]=(long)p3;
  }
  if (DEBUGLEVEL)
  { 
    fprintferr("temps generateurs du groupe de classes: ");
    fprintferr("%ld\n",timer2()); flusherr();
  }
  tetpil=avma; p4=cgetg(6,t_VEC); p4[1]=lcopy(p1);
  mot=cgetg(c+1,t_VEC); p4[2]=(long)mot;
  for (i=1; i<=c; i++) mot[i]=lcopy(gcoeff(met,i,i));
  p4[3]=lcopy(basecl); p4[4]=lcopy(reg); p4[5]=lcopy(c_1);
  free(base); free(vectprime); free(ex);
  free(subbase); free(numbase); free(vperm);
  for (i=1; i<=sizematcol; i++) free(matinit[i]);
  free(matinit); free(mat);
  for (i=1; i<=cardsub; i++) free(tabpowprform[i]);
  free(tabpowprform);
  for (i=1; i<HASHT; i++) freehash(hashtab[i]);
  if (DEBUGLEVEL)
  { 
    fprintferr("temps free finaux: ");
    fprintferr("%ld\n",timer2()); flusherr();
  }
  return gerepile(av,tetpil,p4);
}

static long *
largeprime(long q1, long *ex, long np, long nrho, long cardsubbase,
           long **hashtab)
{
  const long hashv = ((q1&2047)-1)>>1;
  long *pt=hashtab[hashv], cpt=0, i;
  
  while (pt && q1!=pt[1]) { pt=(long*)(*pt); cpt++; }

  if (!pt)
  {
    pt = (long*) xmalloc((cardsubbase+4)<<TWOPOTBYTES_IN_LONG);
    pt[1]=q1;
    for (i=2; i<cardsubbase+2; i++) pt[i]=ex[i-1];
    pt[cardsubbase+2]=np; pt[cardsubbase+3]=nrho;
    pt[0]=cpt ? (long)hashtab[hashv] : 0; hashtab[hashv]=pt;
    return NULL;
  }

  for(i=2; i<cardsubbase+2; i++)
    if (pt[i]!=ex[i-1]) return pt;

  return (pt[i]==np) ? NULL : pt;
}

static GEN
comprealform3(GEN x, GEN y, qf_data *data)
{
  long av,tetpil;
  GEN s,n,d,d1,x1,x2,y1,y2,v1,v2,b3,c3,m,z,p1,r;

  av=avma; s=shifti(addii((GEN)x[2],(GEN)y[2]),-1); n=subii((GEN)y[2],s);
  d=bezout((GEN)y[1],(GEN)x[1],&y1,&x1); d1=bezout(s,d,&x2,&y2);
  v1=divii((GEN)x[1],d1); v2=divii((GEN)y[1],d1);
  m=addii(mulii(mulii(y1,y2),n),mulii((GEN)y[3],x2)); setsigne(m,-signe(m));
  r=modii(m,v1); b3=shifti((p1=mulii(v2,r)),1);
  c3=addii(mulii((GEN)y[3],d1),mulii(r,addii((GEN)y[2],p1)));

  z=cgetg(4,t_VEC);
  z[1]=lmulii(v1,v2); z[2]=laddii((GEN)y[2],b3); z[3]=ldivii(c3,v1);

  tetpil=avma; return gerepile(av,tetpil,redrealform3(z,data));
}

static GEN
comprealform5(GEN x, GEN y, qf_data *data)
{
  long av,tetpil,ss;
  GEN s,n,d,d1,x1,x2,y1,y2,v1,v2,b3,c3,m,z,p1,r;

  av=avma; s=shifti(addii((GEN)x[2],(GEN)y[2]),-1); n=subii((GEN)y[2],s);
  d=bezout((GEN)y[1],(GEN)x[1],&y1,&x1); d1=bezout(s,d,&x2,&y2);
  v1=divii((GEN)x[1],d1); v2=divii((GEN)y[1],d1);
  m=addii(mulii(mulii(y1,y2),n),mulii((GEN)y[3],x2)); setsigne(m,-signe(m));
  r=modii(m,v1); b3=shifti((p1=mulii(v2,r)),1);
  c3=addii(mulii((GEN)y[3],d1),mulii(r,addii((GEN)y[2],p1)));

  z=cgetg(6,t_VEC);
  z[1]=lmulii(v1,v2); z[2]=laddii((GEN)y[2],b3); z[3]=ldivii(c3,v1);
  
  z[5]=lmulrr((GEN)x[5],(GEN)y[5]);
  ss=expo(z[5]);
  if (ss>=EXP220)
  { 
    z[4]=laddii(addsi(1,(GEN)x[4]),(GEN)y[4]);
    setexpo(z[5],ss-EXP220);
  }
  else z[4]=laddii((GEN)x[4],(GEN)y[4]);

  tetpil=avma; return gerepile(av,tetpil,redrealform5(z,data));
}

static GEN
rhorealform3(GEN x, qf_data *data)
{
  static long y[] = { evaltyp(t_VEC) | evallg(4), 0,0,0 };
  long av=avma,tetpil,s;
  GEN p1,nn;
  GEN D=data->D;
  GEN isqrtD=data->isqrtD;

  y[1]=x[3]; s=signe(y[1]); setsigne(y[1],1);
  nn = (cmpii(isqrtD,(GEN)y[1]) >= 0) ? isqrtD : (GEN) y[1];
  p1=shifti((GEN)y[1],1);
  nn=divii(addii(nn,(GEN)x[2]), p1);
  y[2]=lsubii(mulii(nn,p1),(GEN)x[2]);

  setsigne(y[1],s);
  p1=shifti(subii(sqri((GEN)y[2]),D),-2);
  y[3]=ldivii(p1,(GEN)y[1]);

  tetpil=avma; return gerepile(av,tetpil,gcopy(y));
}

static GEN
rhorealform5(GEN x, qf_data *data)
{
  static long y[] = { evaltyp(t_VEC) | evallg(6), 0,0,0,0,0 };
  long av=avma,tetpil,s,ss;
  GEN p1,nn;
  GEN D = data->D;
  GEN isqrtD = data->isqrtD;
  GEN sqrtD = data->sqrtD;

  y[1]=x[3]; s=signe(y[1]); setsigne(y[1],1);
  nn = (cmpii(isqrtD,(GEN)y[1]) >= 0) ? isqrtD : (GEN) y[1];
  p1=shifti((GEN)y[1],1);
  nn=divii(addii(nn,(GEN)x[2]), p1);
  y[2]=lsubii(mulii(nn,p1),(GEN)x[2]);

  setsigne(y[1],s);
  p1=shifti(subii(sqri((GEN)y[2]),D),-2);
  y[3]=ldivii(p1,(GEN)y[1]);

  y[5]=lmulrr(divrr(addir((GEN)x[2],sqrtD),subir((GEN)x[2],sqrtD)),(GEN)x[5]);
  ss=expo(y[5]);
  if (ss>=EXP220)
  { 
    y[4]=laddsi(1,(GEN)x[4]);
    y[5]=lshiftr((GEN)y[5],-EXP220);
  }
  else y[4]=x[4];

  tetpil=avma; return gerepile(av,tetpil,gcopy(y));
}

static GEN
redrealform_proto(GEN x, qf_data *data, GEN (red)(GEN,qf_data *))
{
  long av=avma,tetpil;
  long sens = data->sens;
  GEN isqrtD = data->isqrtD;

  for(;;)
  {
    if (signe(x[2]) > 0 && cmpii((GEN)x[2],isqrtD) <= 0)
    {
      GEN p1 = subii(isqrtD, shifti(absi((GEN)x[1]),1));

      if (signe(p1)<0)
      {
	setsigne(p1,1);
	if (cmpii((GEN)x[2],p1) >= 0) break;
      }
      else if (cmpii((GEN)x[2],p1) > 0) break;
    }
    x=red(x,data);
  }

  if (signe(x[1])<0)
  {
    if (sens || !signe(addii((GEN)x[1],(GEN)x[3])))
    { 
      tetpil=avma;
      return gerepile(av,tetpil,red(x,data));
    }
    setsigne(x[1],1); setsigne(x[3],-1);
  }
  tetpil=avma; return gerepile(av,tetpil,gcopy(x));
}

static GEN
redrealform(GEN x, qf_data *data)
{
  static long y[] = { evaltyp(t_VEC) | evallg(6), 0,0,0,0,0 };
  
  y[1]=x[1]; y[2]=x[2]; y[3]=x[3]; y[4]=zero;
  y[5]=lgetr(data->precreg); affsr(1,(GEN)y[5]);
  return redrealform5(y,data);
}

static GEN
initializeform3(long *ex, GEN **tabform, long cardtab, qf_data *data)
{
  long av = avma, i;
  GEN form = tabform[1][ex[1]];

  for (i=2; i<=cardtab; i++)
    form=comprealform3(form,tabform[i][ex[i]],data);

  i=avma; return gerepile(av,i,gcopy(form));
}

static GEN
initializeform5(long *ex, GEN **tabform, long cardtab, qf_data *data)
{
  long av = avma, i;
  GEN form = tabform[1][ex[1]];

  for (i=2; i<=cardtab; i++)
    form=comprealform5(form,tabform[i][ex[i]],data);

  i=avma; return gerepile(av,i,gcopy(form));
}

#define LIMP 30000
/* L-function */
static GEN
lfunc(GEN D)
{
  GEN y;
  long av=avma,tetpil,prime=0;
  byteptr p=diffptr;

  prime=*p++; affsr(1,y=cgetr(DEFAULTPREC));
  do
  {
    if (!*p) err(primer1);
    y=mulsr(prime,divrs(y,prime-krogs(D,prime)));
    prime+=*p++;
  }
  while (prime<=LIMP);
  tetpil=avma; return gerepile(av,tetpil,gcopy(y));
}

GEN
quadclassunit0(GEN x, long flag, GEN data, long prec)
{
  static long court[] = {evaltyp(t_INT) | evallg(3),0,0};
  static long doub[] = {evaltyp(t_REAL) | evallg(4),0,0,0};
  long lx=lg(data),av=avma,s=signe(x);
  GEN gcbach,gcbach2,gRELSUP,gsens,p1;

  if (gcmp0(data)) lx=1;
  else if (typ(data)!=t_VEC)
    err(talker,"incorrect parameters in classgroupquad");
  if (typ(x)!=t_INT) err(typeer,"classgroupquad");
  if (!s) err(talker,"zero discriminant in classgroupquad");

  affsi(5,court); av=avma; p1=dbltor(0.1); affrr(p1,doub); avma=av;
  gcbach=(lx>1)?(GEN)data[1]:doub;
  gcbach2=(lx>2)?(GEN)data[2]:doub;
  gRELSUP=(lx>3)?(GEN)data[3]:court;
  if (s<0) return buchimag(x,gcbach,gcbach2,gRELSUP);
  gsens=flag?gun:gzero;
  return buchreal(x,gsens,gcbach,gcbach2,gRELSUP,prec);
}
