/**************************************************************/
/*                                                            */
/*    Galois group for degree between 8 and 11 (included)     */
/*                                                            */
/**************************************************************/
/* $Id: galois.c,v 1.926.0.29 1997/06/29 15:38:39 karim Exp karim $ */
#include "genpari.h"

#define NMAX 11 /* maximum degree */

typedef char *PERMUTATION;
typedef char **GROUP;
typedef char **RESOLVANTE;

static long closure8(),galoisprim8(),galoisimpodd8(),galoisimpeven8();
static long closure9(),galoisprim9(),galoisimpodd9(),galoisimpeven9();
static long closure10(),galoisprim10(),galoisimpodd10(),galoisimpeven10();
static long closure11(),galoisodd11(),galoiseven11();
static long galoismodulo8(),galoismodulo9(),galoismodulo10(),galoismodulo11();
static PERMUTATION data8(),data9(),data10(),data11();

static GEN EPS;
static long N,PRMAX,TSCHMAX,coeff[9][10];
static char SID[] = { 0,1,2,3,4,5,6,7,8,9,10,11 };
static char* chaincoset  = GPDATADIR"/COS";
static char* chainresolv = GPDATADIR"/RES";

GEN
galoisbig(GEN pol, long prec)
{
  long av0 = avma, tetpil,ga,precis,car,nbdig;
  GEN dpol,res;

  static long tab8[]={ 0,8,8,8,8,8,16,16,16,16,16,16,24,24,24,32,32,32,32,32,32,32,32,48,48,56,64,64,64,64,64,64,96,96,96,128,168,168,192,192,192,192,288,336,384,576,576,1152,1344,20160,40320 };
  static long tab9[]={ 0,9,9,18,18,18,27,27,36,36,54,54,54,54,72,72,72,81,108,144,162,162,162,216,324,324,432,504,648,648,648,1296,1512,181440,362880 };
  static long tab10[]={ 0,10,10,20,20,40,50,60,80,100,100,120,120,120,160,160,160,200,200,200,200,200,240,320,320,320,360,400,400,640,720,720,720,800,960,1440,1920,1920,1920,3840,7200,14400,14400,28800,1814400,3628800 };
  static long tab11[]={ 0,11,22,55,110,660,7920,19958400,39916800 };

  precis=prec+8; nbdig=(precis-2)*pariK; EPS=gpuigs(gdeux,-16*(precis-2));
  if (DEBUGLEVEL>=2)
  {
    fprintferr("\n        Computation of the Galois group of a polynomial ");
    fprintferr("of degree <= 11.\n");
    fprintferr("                         (decimal precision = %ld)\n\n",nbdig);
    flusherr();
  }
  N = lgef(pol)-3; dpol=discsr(pol);
  if (DEBUGLEVEL)
  {
    fprintferr("Working with reduced polynomial #1 = "); bruterr(pol,'g',-1);
    fprintferr("\ndiscriminant = "); bruterr(dpol,'g',-1);
    fprintferr("\n"); flusherr();
  }
  car=carreparfait(dpol);
  if (DEBUGLEVEL)
    { fprintferr(car? "EVEN group\n": "ODD group\n"); flusherr(); }

  res=cgetg(4,t_VEC); res[2] = car? un: lnegi(gun);
  PRMAX = precis+5; TSCHMAX = 1;
  switch(N)
  {
    case 8:
      switch (galoismodulo8(pol,dpol,car))
      {
        case 49: res[1]=lstoi(tab8[49]); res[3]=lstoi(49); break;
        case 50: res[1]=lstoi(tab8[50]); res[3]=lstoi(50); break;
        default:
          ga=closure8(pol,car,precis);
          res[3]=lstoi(ga);
          res[1]=lstoi(tab8[ga]);
      }
      break;
    case 9:
      switch (galoismodulo9(pol,dpol,car))
      {
       case 33: res[1]=lstoi(tab9[33]); res[3]=lstoi(33); break;
       case 34: res[1]=lstoi(tab9[34]); res[3]=lstoi(34); break;
       default:
         ga=closure9(pol,car,precis);
         res[3]=lstoi(ga);
         res[1]=lstoi(tab9[ga]);
      }
      break;
    case 10:
      switch (galoismodulo10(pol,dpol,car))
      {
        case 44: res[1]=lstoi(tab10[44]); res[3]=lstoi(44); break;
        case 45: res[1]=lstoi(tab10[45]); res[3]=lstoi(45); break;
        default:
          ga=closure10(pol,car,precis);
          res[3]=lstoi(ga);
          res[1]=lstoi(tab10[ga]);
      }
      break;
    case 11:
      switch (galoismodulo11(pol,dpol,car))
      {
        case 1:
          if (car) { res[1]=lstoi(tab11[7]); res[3]=lstoi(7); }
          else     { res[1]=lstoi(tab11[8]); res[3]=lstoi(8); }
          break;
        default:
          ga=closure11(pol,car,precis);
          res[3]=lstoi(ga);
          res[1]=lstoi(tab11[ga]);
      }
      break;
    default: err(impl,"galois in degree > 11");
  }
  tetpil=avma; return gerepile(av0,tetpil,gcopy(res));
}

static long par_N, *par_vec;

static void
do_par(long k, long n, long m)
{
  long i;
  
  if (n<=0)
  {
    GEN p1 = cgeti(par_N+1);
    for (i=1; i<k     ; i++) p1[i] = par_vec[i];
    for (   ; i<=par_N; i++) p1[i] = 0;
    return;
  }
  if (n<m) m=n;
  for (i=1; i<=m; i++)
  {
    par_vec[k] = i;
    do_par(k+1, n-i, i);
  }
}

/* compute the partitions of m. T[0][0] = p(m) */
static long **
partitions(long n)
{
  long av,av1,i, j = 1, l = n+1; 
  GEN T;

  par_vec = cgeti(l); par_N = n;
  l = l*sizeof(long);
  av = avma; do_par(1,n,n); av1 = avma;
  T = cgeti((av-av1)/l + 1);
  for (i=av-l; i>=av1; i-=l) T[j++]=i;

  if (DEBUGLEVEL)
  {
    fprintferr("Partitions of %ld: p(%ld) = %ld\n",n,n,j-1);
    for (i=1; i<j; i++)
    {
      fprintferr("i = %ld: ",i); 
      for (l=1; l<=n; l++)
        fprintferr("%ld ",((long**)T)[i][l]);
      fprintferr("\n"); flusherr();
    }
  }
  T[0] = lgeti(1); ((long**)T)[0][0] = j-1;
  return (long**)T;
}

/* affect to the permutation x the N arguments that follow */
static void
_aff(char *x,...)
{
  va_list args;
  long i;

  va_start(args,x);
  for (i=1; i<=N; i++) x[i] = va_arg(args,char);
  va_end(args);
}

/* return an array of length |len| from the arguments (for galoismodulo) */
static GEN
_affl(long len,...)
{
  va_list args;
  long i, l = labs(len);
  GEN x = cgeti(l);

  va_start(args,len); x[0] = len;
  for (i=1; i<=len; i++) x[i] = va_arg(args,long);
  va_end(args); return x;
}

/* create a permutation with the N arguments of the function */
static PERMUTATION
_cr(char a,...)
{
  PERMUTATION x = xmalloc(N+1);
  va_list args;
  long i;

  va_start(args, a); x[0] = N; x[1] = a;
  for (i=2; i<=N; i++) x[i] = va_arg(args,char);
  va_end(args); return x;
}

static PERMUTATION
permmul(PERMUTATION s1, PERMUTATION s2)
{
  long i, n1 = s1[0];
  PERMUTATION s3 = xmalloc(n1+1);
  for (i=1; i<=n1; i++) s3[i]=s1[s2[i]];
  s3[0]=n1; return s3;
}

static void
printperm(PERMUTATION perm)
{
  long i, n = perm[0]; 
  fprintferr("("); 
  for (i=1; i<=n; i++) fprintferr(" %d",perm[i]);
  fprintferr(" )");
}

static void
reverse(char *t)
{
  long i,j;
  for (i=0,j=strlen(t)-1; i<j; i++,j--)
    { char c=t[i]; t[i]=t[j]; t[j]=c; }
}

static void
itoa(long m, char *t)
{
  long i, s = m;
  if (s<0) m=-m; i=0;
  do { t[i++] = m%10+'0'; m/=10; } while (m);
  if (s<0) t[i++]='-'; t[i]='\0'; reverse(t);
}

/* ranger dans l'ordre decroissant (quicksort) */
static void
ranger(long *t, long n)
{
  long tpro,l,r,i,j;

  l=1+n/2; r=n; tpro=t[1];
  for(;;)
  {
    if (l>1) { l--; tpro=t[l]; }
    else
    { 
      tpro=t[r]; t[r]=t[1]; r--;
      if (r==1) { t[1]=tpro; return; }
    }
    i=l;
    for (j=i<<1; j<=r; j<<=1)
    {
      if (j < r && t[j] > t[j+1]) j++;
      if (t[j] >= tpro) break; 
      t[i] = t[j]; i=j;
    }
    t[i]=tpro;
  }
}

/* 0 si t1=t2, -1 si t1<t2, 1 si t1>t2 */
static long
compareupletlong(long *t1,long *t2)
{
  long i;
  for (i=1; i<=N; i++) 
    if (t1[i]!=t2[i]) return (t1[i] < t2[i])? -1: 1;
  return 0; 
}

/* 0 si typ n'est pas un des TYP[i], i si typ=TYP[i] */
static long
numerotyp(long **TYP, long *galtyp)
{
  long i, nb = TYP[0][0];
  for (i=1; i<=nb; i++) 
    if (!compareupletlong(galtyp,TYP[i])) return i;
  err(talker,"non consistent type in numerotyp()");
  return 0; /* not reached */
}

static long
rayergrouppair(long **GR, long num, long *gr, long nbgr)
{
  long i,j,nb, r = 0;

  for (i=1; i<nbgr; i++)
    if (GR[i][0] < 0 && gr[i])
    {
      nb = -GR[i][0];
      for (j=1; j<=nb; j++)
        if (num==GR[i][j]) break;
      if (j>nb) { gr[i]=0; r++; }
    }
  return r;
}

static long
rayergroupimpair(long **GR, long num, long *gr, long nbgr)
{
  long i,j,nb, r = 0;

  for (i=1; i<nbgr; i++)
    if (GR[i][0] > 0 && gr[i])
    {
      nb = GR[i][0]; 
      for (j=1; j<=nb; j++) 
        if (num==GR[i][j]) break;
      if (j>nb) { gr[i]=0; r++; }
    }
  return r;
}

static long
suffprec(GEN z)
{
  long s,t;

  if (typ(z)==t_COMPLEX)
  { 
    s=suffprec((GEN)z[1]);
    t=suffprec((GEN)z[2]); if (t>s) s=t;
  }
  else if (signe(z)==0) 
    s = (expo(z)+101) >> TWOPOTBITS_IN_LONG; 
  else 
    s = ((expo(z)+165) >> TWOPOTBITS_IN_LONG) - lg(z);
  return s;
}

static void 
preci(GEN *r, long p)
{
  GEN x;
  long d,i;

  if (p>PRMAX) err(talker,"too large precision in preci()");
  for (d=0; d<TSCHMAX; d++) for (i=1; i<=N; i++)
  { 
    x = (GEN) r[d][i];
    if (typ(x)==t_COMPLEX) { setlg(x[1],p); setlg(x[2],p); } else setlg(x,p);
  }
}

static void
rearrange(GEN *r, PERMUTATION s)
{
  long z[NMAX+1],i,j;

  for (i=0; i<TSCHMAX; i++)
  { 
    for (j=1; j<=N; j++) z[j]=r[i][s[j]];
    for (j=1; j<=N; j++) r[i][j]=z[j];
  }
}

static void
rangeroots(GEN newrr, GEN oldrr, long pr)
{
  long av0,i,j,k,z[NMAX+1];
  char s[NMAX+1],t[NMAX+1];
  GEN diff,diff0;

  av0=avma; for (i=1; i<=N; i++) t[i]=1;
  for (i=1; i<=N; i++)
  {
    diff0=gun;
    for (j=1; j<=N; j++)
      if (t[j])
      {
        diff = gabs(gsub((GEN)oldrr[i], (GEN)newrr[j]),pr);
        if (gcmp(diff,diff0) < 0) { diff0=diff; k=j; }
      }
    s[i]=k; t[k]=0;
  }
  avma=av0;
  for (i=1; i<=N; i++) z[i]=newrr[s[i]];
  for (i=1; i<=N; i++) newrr[i]=z[i];
}

static void
new_pol(GEN *r, long *a, long d)
{
  long av = avma,i,j;
  GEN x, p1;
  for (i=1; i<=N; i++)
  { 
    x = gun; p1 = (GEN)r[0][i];
    for (j=0; j<=d; j++) x = gaddsg(a[j], gmul(p1,x)); 
    x = gerepileupto(av,x); r[d][i]=(long)x; av = avma;
  }
}

static void
moreprec(GEN po, GEN *r, long pr, long sp)
{
  GEN rrpro;
  long d;

  if (sp<5) sp=5;
  PRMAX += sp; rrpro = roots(po,PRMAX);
  if (lg(rrpro)-1 != N)
    err(talker,"incompatible number of roots in moreprec()");
  rangeroots(rrpro,*r,pr); *r=rrpro;
  for (d=1; d<TSCHMAX; d++) new_pol(r,coeff[d],d);
  preci(r,pr);
}

static void
setcardgroup(GROUP x, long n) { ((long*)x[0])[1]=n; }

static long
getcardgroup(GROUP x) { return (long) ((long*)x[0])[1]; }

static void
desallocgroup(GROUP gr)
{
  long i = getcardgroup(gr);
  for ( ; i>=0; i--)
    if (gr[i] != SID) free(gr[i]);
  free(gr);
}

static GROUP
allocgroup(long n, long cardgr)
{
  long i;
  GROUP gr = (char**)xmalloc((cardgr+1)*sizeof(char*));

  gr[0] = xmalloc(2); gr[0][0]=(char)n; setcardgroup(gr,cardgr);
  for (i=1; i<=cardgr; i++) { gr[i] = xmalloc(n+1); gr[i][0]=(char)n; }
  return gr;
}

static void
desallocresolv(RESOLVANTE cr, long nbcr)
{
  long i;
  for (i=1; i<=nbcr; i++) free(cr[i]); free(cr);
}

static RESOLVANTE
allocresolv(long nbcr, long lgcr)
{
  long i;
  RESOLVANTE cr = (char**)xmalloc((nbcr+1)*sizeof(char*));
  for (i=1; i<=nbcr; i++) cr[i] = xmalloc(lgcr+1);
  return cr;
}

static char
bin(char c)
{
  if (c>='0' && c<='9') c=c-'0';
  else if (c>='A' && c<='Z') c=c-'A'+10;
  else if (c>='a' && c<='z') c=c-'a'+36;
  else err(talker,"incorrect value in bin()");
  return c;
}

static char *
name(char *pre, long n, long n1, long n2, long no)
{
  static char chn[128];
  char ch[5];
  
  strcpy(chn,pre);
  itoa(n,ch);  strcat(chn,ch); strcat(chn,"_");
  itoa(n1,ch); strcat(chn,ch); strcat(chn,"_");
  itoa(n2,ch); strcat(chn,ch);
  if (no) { strcat(chn,"_"); itoa(no,ch); strcat(chn,ch); }
  return chn;
}

static GROUP
lirecoset(long n1, long n2, long n)
{
  GROUP gr;
  char c, ch[8];
  long i,j,no,m,cardgr,fd;

  if (n<11 || n1<8)
  {
    fd = open(name(chaincoset, n, n1, n2, 0), 0);
    if (fd == -1) 
      err(talker,"galois files not available in this version, sorry");
    read(fd,&c,1); m=(long)bin(c); read(fd,&c,1);
    read(fd,ch,6); cardgr=atol(ch);
    gr=allocgroup(m,cardgr);
    for (i=1; i<=cardgr; i++) for (j=1; j<=m; j++)
      { read(fd,&c,1); gr[i][j]=bin(c); }
    close(fd); return gr;
  }
  gr=allocgroup(n,362880);
  for (no=1; no<9; no++)
  {
    fd = open(name(chaincoset, n, n1, n2, no), 0);
    if (fd == -1) 
      err(talker,"galois files not available in this version, sorry");
    read(fd,ch,8); m=45360*(no-1);
    for (i=1; i<=45360; i++) for (j=1; j<=11; j++)
      { read(fd,&c,1); gr[m+i][j]=bin(c); }
    close(fd);
  }
  return gr;
}

static RESOLVANTE
lireresolv(long n1, long n2, long n, long *nv, long *nm)
{
  RESOLVANTE b;
  char c, ch[5];
  long fd,i,j;

  fd = open(name(chainresolv, n, n1, n2, 0), 0);
  if (fd == -1)
    err(talker,"galois files not available in this version, sorry");
  read(fd,ch,5); *nm=atol(ch);
  read(fd,ch,3); *nv=atol(ch);
  b=allocresolv(*nm,*nv);
  for (i=1; i<=*nm; i++) for (j=1; j<=*nv; j++)
    { read(fd,&c,1); b[i][j]=bin(c); }
  close(fd); return b;
}

static GEN
monomial(GEN rr, PERMUTATION tt, PERMUTATION bb, long nbv)
{
  long i;
  GEN p1 = (GEN)rr[tt[bb[1]]];

  for (i=2; i<=nbv; i++) p1=gmul(p1, (GEN)rr[tt[bb[i]]]);
  return p1;
}

static GEN
gpolynomial(GEN rr, PERMUTATION tt, RESOLVANTE aa, long nbm, long nbv)
{
  long av = avma,tetpil,i;
  GEN p2, p1 = monomial(rr,tt,aa[1],nbv);
  
  for (i=2; i<=nbm; i++)
    { p2=monomial(rr,tt,aa[i],nbv); tetpil=avma; p1=gadd(p1,p2); }
  return gerepile(av,tetpil,p1);
}

static GEN
gpoly(GEN rr, PERMUTATION tt, long n1, long n2)
{
  long av,tetpil,i,j;
  GEN p1,p2,p3,p4,z[6];
  av=avma;
  if (N==8)
  {
    if (n1==47 && n2==46)
    {
      p1=gsub((GEN)rr[tt[3]],(GEN)rr[tt[4]]);
      for (i=1; i<3; i++) for (j=i+1; j<5; j++)
        p1 = gmul(p1,gsub((GEN)rr[tt[i]],(GEN)rr[tt[j]]));
      for (i=5; i<8; i++) for (j=i+1; j<9; j++)
        p1 = gmul(p1,gsub((GEN)rr[tt[i]],(GEN)rr[tt[j]]));
      p2=(GEN)rr[tt[1]];
      for (i=2; i<5; i++) p2=gadd(p2,(GEN)rr[tt[i]]);
      for (i=5; i<9; i++) p2=gsub(p2,(GEN)rr[tt[i]]);
    }
    else if (n1==44 && n2==40)
    {
      p1 = gsub((GEN)rr[tt[1]],(GEN)rr[tt[2]]);
      for (i=2; i<5; i++)
        p1 = gmul(p1,gsub((GEN)rr[tt[2*i-1]],(GEN)rr[tt[2*i]]));
      for (i=1; i<5; i++)
        z[i] = gadd((GEN)rr[tt[2*i-1]],(GEN)rr[tt[2*i]]);
      p2=gsub(z[3],z[4]);
      for (i=1; i<3; i++) for (j=i+1; j<5; j++) 
        p2 = gmul(p2,gsub(z[i],z[j]));
    }
    tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
  }

  if (N==9)
  {
    if (n1==31 && n2==29)
    {
      p1=gsub((GEN)rr[tt[2]],(GEN)rr[tt[3]]);
      for (j=2; j<4; j++)
        p1 = gmul(p1,gsub((GEN)rr[tt[1]],(GEN)rr[tt[j]]));
      for (i=4; i<6; i++) for (j=i+1; j<7; j++)
        p1 = gmul(p1,gsub((GEN)rr[tt[i]],(GEN)rr[tt[j]]));
      p2 = gsub((GEN)rr[tt[8]],(GEN)rr[tt[9]]);
      for (j=8; j<10; j++)
        p2 = gmul(p2,gsub((GEN)rr[tt[7]],(GEN)rr[tt[j]]));
    }
    else if ((n1==34 && n2==31) || (n1=33 && n2==30))
    {
      p1=(GEN)rr[tt[1]]; for (i=2; i<4; i++) p1=gadd(p1,(GEN)rr[tt[i]]);
      p2=(GEN)rr[tt[4]]; for (i=5; i<7; i++) p2=gadd(p2,(GEN)rr[tt[i]]);
      p1=gmul(p1,p2); p2=(GEN)rr[tt[7]]; 
      for (i=8; i<10; i++) p2=gadd(p2,(GEN)rr[tt[i]]);
    }
    tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
  }

  if (N==10)
  {
    if ((n1==45 && n2==43) || (n1==44 && n2==42))
    {
      p1=(GEN)rr[tt[1]]; for (i=2; i<6; i++) p1=gadd(p1,(GEN)rr[tt[i]]);
      p2=(GEN)rr[tt[6]]; for (i=7; i<11; i++) p2=gadd(p2,(GEN)rr[tt[i]]);
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
    }
    else if ((n1==45 && n2==39) || (n1==44 && n2==37))
    {
      p1 = gadd((GEN)rr[tt[1]],(GEN)rr[tt[2]]);
      for (i=2; i<5; i++)
        p1 = gmul(p1,gadd((GEN)rr[tt[2*i-1]],(GEN)rr[tt[2*i]]));
      p2 = gadd((GEN)rr[tt[9]],(GEN)rr[tt[10]]);
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
    }
    else if ((n1==43 && n2==41) || (n1==33 && n2==27))
    {
      p1=gsub((GEN)rr[tt[4]],(GEN)rr[tt[5]]);
      for (i=1; i<4; i++)
       for (j=i+1; j<6; j++) p1=gmul(p1,gsub((GEN)rr[tt[i]],(GEN)rr[tt[j]]));
      p2=gsub((GEN)rr[tt[9]],(GEN)rr[tt[10]]);
      for (i=6; i<9; i++)
       for (j=i+1; j<11; j++) p2=gmul(p2,gsub((GEN)rr[tt[i]],(GEN)rr[tt[j]]));
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
    }
    else if ((n1==43 && n2==33) || (n1==42 && n2==28) || (n1==41 && n2==27)
          || (n1==40 && n2==21))
    {
      p2=gadd((GEN)rr[tt[2]],(GEN)rr[tt[5]]);
      p2=gsub(p2,gadd((GEN)rr[tt[3]],(GEN)rr[tt[4]]));
      p1=gmul(p2,(GEN)rr[tt[1]]);
      p2=gsub((GEN)rr[tt[3]],gadd((GEN)rr[tt[4]],(GEN)rr[tt[5]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[2]]));
      p2=gsub((GEN)rr[tt[4]],(GEN)rr[tt[5]]);
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[3]]));
      p1=gadd(p1,gmul((GEN)rr[tt[4]],(GEN)rr[tt[5]]));
 
      p1=gsqr(p1); p4=gadd((GEN)rr[tt[7]],(GEN)rr[tt[10]]);
      p4=gsub(p4,gadd((GEN)rr[tt[8]],(GEN)rr[tt[9]]));
      p3=gmul(p4,(GEN)rr[tt[6]]);
      p4=gsub((GEN)rr[tt[8]],gadd((GEN)rr[tt[9]],(GEN)rr[tt[10]]));
      p3=gadd(p3,gmul(p4,(GEN)rr[tt[7]]));
      p4=gsub((GEN)rr[tt[9]],(GEN)rr[tt[10]]);
      p3=gadd(p3,gmul(p4,(GEN)rr[tt[8]]));
      p3=gadd(p3,gmul((GEN)rr[tt[9]],(GEN)rr[tt[10]]));
      p3=gsqr(p3); tetpil=avma; return gerepile(av,tetpil,gadd(p1,p3));
    }
    else if (n1==41 && n2==40)
    {
      p1=gsub((GEN)rr[tt[4]],(GEN)rr[tt[5]]);
      for (i=1; i<4; i++) for (j=i+1; j<6; j++)
        p1 = gmul(p1,gsub((GEN)rr[tt[i]],(GEN)rr[tt[j]]));
      p2=gsub((GEN)rr[tt[9]],(GEN)rr[tt[10]]);
      for (i=6; i<9; i++) for (j=i+1; j<11; j++)
        p2 = gmul(p2,gsub((GEN)rr[tt[i]],(GEN)rr[tt[j]]));
      tetpil=avma; return gerepile(av,tetpil,gadd(p1,p2));
    }
    else if ((n1==41 && n2==22) || (n1==40 && n2==11) || (n1==17 && n2==5)
            || (n1==10 && n2==4) || (n1==9 && n2==3) || (n1==6 && n2==1))
    {
      p1=gadd((GEN)rr[tt[1]],(GEN)rr[tt[6]]);
      for (i=2; i<5; i++) p1=gmul(p1,gadd((GEN)rr[tt[i]],(GEN)rr[tt[i+5]]));
      p2=gadd((GEN)rr[tt[5]],(GEN)rr[tt[10]]);
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
    }
    else if ((n1==39 && n2==38) || (n1==29 && n2==25))
    {
      p1=gsub((GEN)rr[tt[1]],(GEN)rr[tt[2]]);
      for (i=2; i<6; i++) p1=gmul(p1,gsub((GEN)rr[tt[2*i-1]],(GEN)rr[tt[2*i]]));
      for (i=1; i<6; i++) z[i]=gadd((GEN)rr[tt[2*i-1]],(GEN)rr[tt[2*i]]);
      p2=gsub(z[4],z[5]);
      for (i=1; i<4; i++) for (j=i+1; j<6; j++) p2=gmul(p2,gsub(z[i],z[j]));
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
    }
    else if ((n1==39 && n2==36) || (n1==37 && n2==34) || (n1==29 && n2==23)
          || (n1==24 && n2==15))
    {
      for (i=1; i<6; i++) z[i]=gadd((GEN)rr[tt[2*i-1]],(GEN)rr[tt[2*i]]); 	
      p1=gsub(z[4],z[5]); p2=gmul(gsub(z[3],z[4]),gsub(z[3],z[5]));
      for (i=1; i<3; i++) for (j=i+1; j<6; j++) p2=gmul(p2,gsub(z[i],z[j])); 	
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
    }
    else if ((n1==39 && n2==29) || (n1==38 && n2==25) || (n1==37 && n2==24)
          || (n1==36 && n2==23) || (n1==34 && n2==15))
    {
      for (i=1; i<6; i++) z[i]=gadd((GEN)rr[tt[2*i-1]],(GEN)rr[tt[2*i]]);
      p2=gadd(z[2],z[5]); p2=gsub(p2,gadd(z[3],z[4])); p1=gmul(p2,z[1]);
      p2=gsub(z[3],gadd(z[4],z[5])); p1=gadd(p1,gmul(p2,z[2]));
      p2=gsub(z[4],z[5]); p1=gadd(p1,gmul(p2,z[3]));
      p1=gadd(p1,gmul(z[4],z[5]));
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p1));
    }
    else if ((n1==39 && n2==22) || (n1==38 && n2==12) || (n1==36 && n2==11)
          || (n1==29 && n2== 5) || (n1==25 && n2== 4) || (n1==23 && n2== 3)
          || (n1==16 && n2== 2) || (n1==14 && n2== 1))
    {
      p1=(GEN)rr[tt[1]]; for (i=2; i<6; i++) p1=gadd(p1,(GEN)rr[tt[2*i-1]]);
      p2=(GEN)rr[tt[2]]; for (i=2; i<6; i++) p2=gadd(p2,(GEN)rr[tt[2*i]]);
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2)); 	
    }
    else if (n1==28 && n2==18)
    {
      p2=gsub((GEN)rr[tt[1]],gadd((GEN)rr[tt[2]],(GEN)rr[tt[5]]));
      p2=gmul(p2,gsub((GEN)rr[tt[2]],(GEN)rr[tt[5]]));
      p1=gmul(p2,(GEN)rr[tt[1]]);
      p2=gsub((GEN)rr[tt[3]],gadd((GEN)rr[tt[2]],(GEN)rr[tt[4]]));
      p2=gmul(p2,gsub((GEN)rr[tt[4]],(GEN)rr[tt[2]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[3]]));
      p2=gmul((GEN)rr[tt[5]],gsub((GEN)rr[tt[4]],(GEN)rr[tt[5]]));
      z[1]=gadd(p1,gmul(p2,(GEN)rr[tt[4]]));
      p2=gsub((GEN)rr[tt[1]],gadd((GEN)rr[tt[3]],(GEN)rr[tt[4]]));
      p2=gmul(p2,gsub((GEN)rr[tt[3]],(GEN)rr[tt[4]]));
      p1=gmul(p2,(GEN)rr[tt[1]]);
      p2=gsub((GEN)rr[tt[5]],gadd((GEN)rr[tt[3]],(GEN)rr[tt[2]]));
      p2=gmul(p2,gsub((GEN)rr[tt[2]],(GEN)rr[tt[3]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[5]]));
      p2=gmul((GEN)rr[tt[4]],gsub((GEN)rr[tt[2]],(GEN)rr[tt[4]]));
      z[2]=gadd(p1,gmul(p2,(GEN)rr[tt[2]]));
      p2=gsub((GEN)rr[tt[6]],gadd((GEN)rr[tt[7]],(GEN)rr[tt[10]]));
      p2=gmul(p2,gsub((GEN)rr[tt[7]],(GEN)rr[tt[10]]));
      p1=gmul(p2,(GEN)rr[tt[6]]);
      p2=gsub((GEN)rr[tt[8]],gadd((GEN)rr[tt[7]],(GEN)rr[tt[9]]));
      p2=gmul(p2,gsub((GEN)rr[tt[9]],(GEN)rr[tt[7]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[8]]));
      p2=gmul((GEN)rr[tt[10]],gsub((GEN)rr[tt[9]],(GEN)rr[tt[10]]));
      z[3]=gadd(p1,gmul(p2,(GEN)rr[tt[9]]));
      p2=gsub((GEN)rr[tt[6]],gadd((GEN)rr[tt[8]],(GEN)rr[tt[9]]));
      p2=gmul(p2,gsub((GEN)rr[tt[8]],(GEN)rr[tt[9]]));
      p1=gmul(p2,(GEN)rr[tt[6]]);
      p2=gsub((GEN)rr[tt[10]],gadd((GEN)rr[tt[8]],(GEN)rr[tt[7]]));
      p2=gmul(p2,gsub((GEN)rr[tt[7]],(GEN)rr[tt[8]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[10]]));
      p2=gmul((GEN)rr[tt[9]],gsub((GEN)rr[tt[7]],(GEN)rr[tt[9]]));
      z[4]=gadd(p1,gmul(p2,(GEN)rr[tt[7]]));
      p1=gmul(z[1],gsub(z[3],z[4]));
      p2=gmul(z[2],gadd(z[3],z[4]));
      tetpil=avma; return gerepile(av,tetpil,gadd(p1,p2));
    }
    else if (n1==27 && n2==20)
    {
      p2=gsub((GEN)rr[tt[1]],gadd((GEN)rr[tt[2]],(GEN)rr[tt[5]]));
      p2=gmul(p2,gsub((GEN)rr[tt[2]],(GEN)rr[tt[5]]));
      p1=gmul(p2,(GEN)rr[tt[1]]);
      p2=gsub((GEN)rr[tt[3]],gadd((GEN)rr[tt[2]],(GEN)rr[tt[4]]));
      p2=gmul(p2,gsub((GEN)rr[tt[4]],(GEN)rr[tt[2]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[3]]));
      p2=gmul((GEN)rr[tt[5]],gsub((GEN)rr[tt[4]],(GEN)rr[tt[5]]));
      z[1]=gadd(p1,gmul(p2,(GEN)rr[tt[4]]));
      p2=gsub((GEN)rr[tt[1]],gadd((GEN)rr[tt[3]],(GEN)rr[tt[4]]));
      p2=gmul(p2,gsub((GEN)rr[tt[3]],(GEN)rr[tt[4]]));
      p1=gmul(p2,(GEN)rr[tt[1]]);
      p2=gsub((GEN)rr[tt[5]],gadd((GEN)rr[tt[3]],(GEN)rr[tt[2]]));
      p2=gmul(p2,gsub((GEN)rr[tt[2]],(GEN)rr[tt[3]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[5]]));
      p2=gmul((GEN)rr[tt[4]],gsub((GEN)rr[tt[2]],(GEN)rr[tt[4]]));
      z[2]=gadd(p1,gmul(p2,(GEN)rr[tt[2]]));
      p2=gsub((GEN)rr[tt[6]],gadd((GEN)rr[tt[7]],(GEN)rr[tt[10]]));
      p2=gmul(p2,gsub((GEN)rr[tt[7]],(GEN)rr[tt[10]]));
      p1=gmul(p2,(GEN)rr[tt[6]]);
      p2=gsub((GEN)rr[tt[8]],gadd((GEN)rr[tt[7]],(GEN)rr[tt[9]]));
      p2=gmul(p2,gsub((GEN)rr[tt[9]],(GEN)rr[tt[7]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[8]]));
      p2=gmul((GEN)rr[tt[10]],gsub((GEN)rr[tt[9]],(GEN)rr[tt[10]]));
      z[3]=gadd(p1,gmul(p2,(GEN)rr[tt[9]]));
      p2=gsub((GEN)rr[tt[6]],gadd((GEN)rr[tt[8]],(GEN)rr[tt[9]]));
      p2=gmul(p2,gsub((GEN)rr[tt[8]],(GEN)rr[tt[9]]));
      p1=gmul(p2,(GEN)rr[tt[6]]);
      p2=gsub((GEN)rr[tt[10]],gadd((GEN)rr[tt[8]],(GEN)rr[tt[7]]));
      p2=gmul(p2,gsub((GEN)rr[tt[7]],(GEN)rr[tt[8]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[10]]));
      p2=gmul((GEN)rr[tt[9]],gsub((GEN)rr[tt[7]],(GEN)rr[tt[9]]));
      z[4]=gadd(p1,gmul(p2,(GEN)rr[tt[7]]));
      p1=gsub(gmul(z[1],z[3]),gmul(z[2],z[4]));
      p2=(GEN)rr[tt[1]]; for (i=2; i<6; i++)p2=gadd(p2,(GEN)rr[tt[i]]);
      for (i=6; i<11; i++) p2=gsub(p2,(GEN)rr[tt[i]]);
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
    }
    else if (n1==27 && n2==19)
    {
      p2=gsub((GEN)rr[tt[1]],gadd((GEN)rr[tt[2]],(GEN)rr[tt[5]]));
      p2=gmul(p2,gsub((GEN)rr[tt[2]],(GEN)rr[tt[5]]));
      p1=gmul(p2,(GEN)rr[tt[1]]);
      p2=gsub((GEN)rr[tt[3]],gadd((GEN)rr[tt[2]],(GEN)rr[tt[4]]));
      p2=gmul(p2,gsub((GEN)rr[tt[4]],(GEN)rr[tt[2]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[3]]));
      p2=gmul((GEN)rr[tt[5]],gsub((GEN)rr[tt[4]],(GEN)rr[tt[5]]));
      z[1]=gadd(p1,gmul(p2,(GEN)rr[tt[4]]));
      p2=gsub((GEN)rr[tt[1]],gadd((GEN)rr[tt[3]],(GEN)rr[tt[4]]));
      p2=gmul(p2,gsub((GEN)rr[tt[3]],(GEN)rr[tt[4]]));
      p1=gmul(p2,(GEN)rr[tt[1]]);
      p2=gsub((GEN)rr[tt[5]],gadd((GEN)rr[tt[3]],(GEN)rr[tt[2]]));
      p2=gmul(p2,gsub((GEN)rr[tt[2]],(GEN)rr[tt[3]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[5]]));
      p2=gmul((GEN)rr[tt[4]],gsub((GEN)rr[tt[2]],(GEN)rr[tt[4]]));
      z[2]=gadd(p1,gmul(p2,(GEN)rr[tt[2]]));
      p2=gsub((GEN)rr[tt[6]],gadd((GEN)rr[tt[7]],(GEN)rr[tt[10]]));
      p2=gmul(p2,gsub((GEN)rr[tt[7]],(GEN)rr[tt[10]]));
      p1=gmul(p2,(GEN)rr[tt[6]]);
      p2=gsub((GEN)rr[tt[8]],gadd((GEN)rr[tt[7]],(GEN)rr[tt[9]]));
      p2=gmul(p2,gsub((GEN)rr[tt[9]],(GEN)rr[tt[7]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[8]]));
      p2=gmul((GEN)rr[tt[10]],gsub((GEN)rr[tt[9]],(GEN)rr[tt[10]]));
      z[3]=gadd(p1,gmul(p2,(GEN)rr[tt[9]]));
      p2=gsub((GEN)rr[tt[6]],gadd((GEN)rr[tt[8]],(GEN)rr[tt[9]]));
      p2=gmul(p2,gsub((GEN)rr[tt[8]],(GEN)rr[tt[9]]));
      p1=gmul(p2,(GEN)rr[tt[6]]);
      p2=gsub((GEN)rr[tt[10]],gadd((GEN)rr[tt[8]],(GEN)rr[tt[7]]));
      p2=gmul(p2,gsub((GEN)rr[tt[7]],(GEN)rr[tt[8]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[10]]));
      p2=gmul((GEN)rr[tt[9]],gsub((GEN)rr[tt[7]],(GEN)rr[tt[9]]));
      z[4]=gadd(p1,gmul(p2,(GEN)rr[tt[7]]));
      p1=gmul(z[1],z[3]); p2=gmul(z[2],z[4]);
      tetpil=avma; return gerepile(av,tetpil,gsub(p1,p2));
    }
    else if ((n1==27 && n2==17) || (n1==21 && n2==9))
    {
      p2=gsub((GEN)rr[tt[1]],gadd((GEN)rr[tt[2]],(GEN)rr[tt[5]]));
      p2=gmul(p2,gsub((GEN)rr[tt[2]],(GEN)rr[tt[5]]));
      p1=gmul(p2,(GEN)rr[tt[1]]);
      p2=gsub((GEN)rr[tt[3]],gadd((GEN)rr[tt[2]],(GEN)rr[tt[4]]));
      p2=gmul(p2,gsub((GEN)rr[tt[4]],(GEN)rr[tt[2]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[3]]));
      p2=gmul((GEN)rr[tt[5]],gsub((GEN)rr[tt[4]],(GEN)rr[tt[5]]));
      z[1]=gadd(p1,gmul(p2,(GEN)rr[tt[4]]));
      p2=gsub((GEN)rr[tt[1]],gadd((GEN)rr[tt[3]],(GEN)rr[tt[4]]));
      p2=gmul(p2,gsub((GEN)rr[tt[3]],(GEN)rr[tt[4]]));
      p1=gmul(p2,(GEN)rr[tt[1]]);
      p2=gsub((GEN)rr[tt[5]],gadd((GEN)rr[tt[3]],(GEN)rr[tt[2]]));
      p2=gmul(p2,gsub((GEN)rr[tt[2]],(GEN)rr[tt[3]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[5]]));
      p2=gmul((GEN)rr[tt[4]],gsub((GEN)rr[tt[2]],(GEN)rr[tt[4]]));
      z[2]=gadd(p1,gmul(p2,(GEN)rr[tt[2]]));
      p2=gsub((GEN)rr[tt[6]],gadd((GEN)rr[tt[7]],(GEN)rr[tt[10]]));
      p2=gmul(p2,gsub((GEN)rr[tt[7]],(GEN)rr[tt[10]]));
      p1=gmul(p2,(GEN)rr[tt[6]]);
      p2=gsub((GEN)rr[tt[8]],gadd((GEN)rr[tt[7]],(GEN)rr[tt[9]]));
      p2=gmul(p2,gsub((GEN)rr[tt[9]],(GEN)rr[tt[7]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[8]]));
      p2=gmul((GEN)rr[tt[10]],gsub((GEN)rr[tt[9]],(GEN)rr[tt[10]]));
      z[3]=gadd(p1,gmul(p2,(GEN)rr[tt[9]]));
      p2=gsub((GEN)rr[tt[6]],gadd((GEN)rr[tt[8]],(GEN)rr[tt[9]]));
      p2=gmul(p2,gsub((GEN)rr[tt[8]],(GEN)rr[tt[9]]));
      p1=gmul(p2,(GEN)rr[tt[6]]);
      p2=gsub((GEN)rr[tt[10]],gadd((GEN)rr[tt[8]],(GEN)rr[tt[7]]));
      p2=gmul(p2,gsub((GEN)rr[tt[7]],(GEN)rr[tt[8]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[10]]));
      p2=gmul((GEN)rr[tt[9]],gsub((GEN)rr[tt[7]],(GEN)rr[tt[9]]));
      z[4]=gadd(p1,gmul(p2,(GEN)rr[tt[7]]));
      p1=gmul(z[1],z[3]); p2=gmul(z[2],z[4]);
      tetpil=avma; return gerepile(av,tetpil,gadd(p1,p2));
    }
    else if (n1==23 && n2==16)
    {
      p1=gsub((GEN)rr[tt[1]],(GEN)rr[tt[2]]);
      for (i=2; i<6; i++) p1=gmul(p1,gsub((GEN)rr[tt[2*i-1]],(GEN)rr[tt[2*i]]));
      for (i=1; i<6; i++) z[i]=gadd((GEN)rr[tt[2*i-1]],(GEN)rr[tt[2*i]]);
      p3=gsub(z[1],gadd(z[2],z[5])); p3=gmul(p3,gsub(z[2],z[5]));
      p2=gmul(p3,z[1]); p3=gsub(z[3],gadd(z[2],z[4]));
      p3=gmul(p3,gsub(z[4],z[2])); p2=gadd(p2,gmul(p3,z[3]));
      p3=gmul(z[5],gsub(z[4],z[5])); p2=gadd(p2,gmul(p3,z[4]));
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
    }
    else if (n1==22 && n2==12)
    {
      p1=gsub((GEN)rr[tt[1]],(GEN)rr[tt[6]]);
      for (i=2; i<6; i++) p1=gmul(p1,gsub((GEN)rr[tt[i]],(GEN)rr[tt[i+5]]));
      for (i=1; i<6; i++) z[i]=gadd((GEN)rr[tt[i]],(GEN)rr[tt[i+5]]);
      p2=gsub(z[4],z[5]);
      for (i=1; i<4; i++) for (j=i+1; j<6; j++) p2=gmul(p2,gsub(z[i],z[j]));
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
    }
    else if ((n1==22 && n2==11) || (n1==5 && n2==3))
    {
      for (i=1; i<6; i++) z[i]=gadd((GEN)rr[tt[i]],(GEN)rr[tt[i+5]]); 	
      p1=gsub(z[4],z[5]); p2=gmul(gsub(z[3],z[4]),gsub(z[3],z[5]));
      for (i=1; i<3; i++) for (j=i+1; j<6; j++) p2=gmul(p2,gsub(z[i],z[j])); 	
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
    }
    else if ((n1==22 && n2==5) || (n1==12 && n2==4) || (n1==11 && n2==3))
    {
      for (i=1; i<6; i++) z[i]=gadd((GEN)rr[tt[i]],(GEN)rr[tt[i+5]]);
      p2=gadd(z[2],z[5]); p2=gsub(p2,gadd(z[3],z[4])); p1=gmul(p2,z[1]);
      p2=gsub(z[3],gadd(z[4],z[5])); p1=gadd(p1,gmul(p2,z[2]));
      p2=gsub(z[4],z[5]);
      p1=gadd(p1,gmul(p2,z[3])); p1=gadd(p1,gmul(z[4],z[5]));
      tetpil=avma; return gerepile(av,tetpil,gmul(p1,p1));
    }
    else if (n1==21 && n2==10)
    {
      p2=gsub((GEN)rr[tt[1]],gadd((GEN)rr[tt[2]],(GEN)rr[tt[5]]));
      p2=gmul(p2,gsub((GEN)rr[tt[2]],(GEN)rr[tt[5]]));
      p1=gmul(p2,(GEN)rr[tt[1]]);
      p2=gsub((GEN)rr[tt[3]],gadd((GEN)rr[tt[2]],(GEN)rr[tt[4]]));
      p2=gmul(p2,gsub((GEN)rr[tt[4]],(GEN)rr[tt[2]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[3]]));
      p2=gmul((GEN)rr[tt[5]],gsub((GEN)rr[tt[4]],(GEN)rr[tt[5]]));
      z[1]=gadd(p1,gmul(p2,(GEN)rr[tt[4]]));
      p2=gsub((GEN)rr[tt[1]],gadd((GEN)rr[tt[3]],(GEN)rr[tt[4]]));
      p2=gmul(p2,gsub((GEN)rr[tt[3]],(GEN)rr[tt[4]]));
      p1=gmul(p2,(GEN)rr[tt[1]]);
      p2=gsub((GEN)rr[tt[5]],gadd((GEN)rr[tt[3]],(GEN)rr[tt[2]]));
      p2=gmul(p2,gsub((GEN)rr[tt[2]],(GEN)rr[tt[3]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[5]]));
      p2=gmul((GEN)rr[tt[4]],gsub((GEN)rr[tt[2]],(GEN)rr[tt[4]]));
      z[2]=gadd(p1,gmul(p2,(GEN)rr[tt[2]]));
      p2=gsub((GEN)rr[tt[6]],gadd((GEN)rr[tt[7]],(GEN)rr[tt[10]]));
      p2=gmul(p2,gsub((GEN)rr[tt[7]],(GEN)rr[tt[10]]));
      p1=gmul(p2,(GEN)rr[tt[6]]);
      p2=gsub((GEN)rr[tt[8]],gadd((GEN)rr[tt[7]],(GEN)rr[tt[9]]));
      p2=gmul(p2,gsub((GEN)rr[tt[9]],(GEN)rr[tt[7]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[8]]));
      p2=gmul((GEN)rr[tt[10]],gsub((GEN)rr[tt[9]],(GEN)rr[tt[10]]));
      z[3]=gadd(p1,gmul(p2,(GEN)rr[tt[9]]));
      p2=gsub((GEN)rr[tt[6]],gadd((GEN)rr[tt[8]],(GEN)rr[tt[9]]));
      p2=gmul(p2,gsub((GEN)rr[tt[8]],(GEN)rr[tt[9]]));
      p1=gmul(p2,(GEN)rr[tt[6]]);
      p2=gsub((GEN)rr[tt[10]],gadd((GEN)rr[tt[8]],(GEN)rr[tt[7]]));
      p2=gmul(p2,gsub((GEN)rr[tt[7]],(GEN)rr[tt[8]]));
      p1=gadd(p1,gmul(p2,(GEN)rr[tt[10]]));
      p2=gmul((GEN)rr[tt[9]],gsub((GEN)rr[tt[7]],(GEN)rr[tt[9]]));
      z[4]=gadd(p1,gmul(p2,(GEN)rr[tt[7]]));
      p1=gmul(z[1],z[4]); p2=gmul(z[2],z[3]);
      tetpil=avma;
      return gerepile(av,tetpil,gsub(p1,p2));
    }
  }
  err(talker,"indefinite invariant polynomial in gpoly()");
  return NULL; /* not reached */
}

static void
tschirn(GEN po, GEN *r, long pr)
{
  long av0 = avma, a[NMAX],v,i,d;
  GEN h,u;

  d = TSCHMAX+1;
  if (d>=N)
    err(talker,"too large degree for Tschirnhausen transformation in tschirn");
  if (DEBUGLEVEL)
  {
    fprintferr("\n$$$$$ Tschirnhausen transformation of degree %ld: $$$$$\n",d);
    flusherr();
  }
  v=varn(po); h=polun[v]; setrand(1);
  do
  { 
    avma = av0;
    for (i=0; i<d; i++)
    { 
      a[i] = ((rand()>>4) & 7) + 1;
      h = gaddsg(a[i],gmul(polx[v],h));
    }
    u=caract(gmodulcp(h,po),v);
  }
  while (lgef(polgcd(u,deriv(u,v))) > 3);
  if (DEBUGLEVEL>2) { bruterr(u,'g',-1); fprintferr("\n"); flusherr(); }
  avma = av0;
  for (i=0; i<d; i++) coeff[d-1][i]=a[i];
  preci(r,PRMAX); r[d-1]=cgetg(N+1,t_VEC);
  new_pol(r,a,d-1); preci(r,pr); TSCHMAX++;
}

static GEN
get_pol(GEN rr, PERMUTATION uu, RESOLVANTE a, long nbm, long nbv)
{
  return a? gpolynomial(rr,uu,a,nbm,nbv): gpoly(rr,uu,nbm,nbv);
}

static GEN
get_pol_perm(PERMUTATION S1, PERMUTATION S2, GEN rr, RESOLVANTE a,
             long nbm, long nbv)
{
  PERMUTATION uu = permmul(S1,S2);
  GEN gpol = get_pol(rr,uu,a,nbm,nbv);
  free(uu); return gpol;
}

static void
dbg_rac(long nri,long nbracint,long numeri[],GEN racint[],long multi[])
{
  long k;
  if (nbracint>nri+1)
    fprintferr("         there are %ld rational integer roots:\n",nbracint-nri);
  else if (nbracint==nri+1)
    fprintferr("         there is 1 rational integer root:\n");
  else 
    fprintferr("         there is no rational integer root.\n");
  for (k=nri+1; k<=nbracint; k++)
  {
    fprintferr("            number%2ld: ",numeri[k]);
    bruterr(racint[k],'g',-1); fprintferr(", order %ld.\n",multi[k]);
  }
  flusherr();
}

static GEN
is_int_pol(GEN gpol, long pr)
{
  GEN p1 = greal(gpol), gpolint = ground(p1);
  if (gcmp(gabs(gimag(gpol),pr),EPS) < 0 &&
      gcmp(gabs(gsub(p1,gpolint),pr),EPS) < 0) return gpolint;
  return NULL;
}

static long
isin_G_H(GEN po, GEN *r, long n1, long n2, long pr)
{
  long nbv,nbm,nbgr,nbcoset,nbracint,nbracmul,lastnbri,lastnbrm;
  long nogr,nocoset,f1,f2,f3,i,j,k,d,nrm,nri,sp,cpr,nbdig;
  long numeri[2521],numerj[2521],lastnum[2521],multi[2521],norac[2521];
  long lastnor[2521],av1,av2,tetpil;
  PERMUTATION s0,uu,vv,ww;
  RESOLVANTE a;
  GROUP ss,tau;
  GEN gpol,gpolint,racint[2521],rr;

  if (DEBUGLEVEL)
  { 
    fprintferr("\n*** Entering in isin_%ld_G_H_(%ld,%ld)\n",N,n1,n2);
    flusherr();
  }
  f1 = 1;
  switch(N)
  { 
    case 8:
      if ((n1==47 && n2==46) || (n1==44 && n2==40)) f1=0;
      s0=data8(n1,n2,&tau); break;
    case 9:
      if ((n1==31 && n2==29) || (n1==34 && n2==31) || (n1==33 && n2==30)) f1=0;
      s0=data9(n1,n2,&tau); break;
    case 10:
      if ((n1==45 && (n2==43||n2==39))
       || (n1==44 && (n2==42||n2==37))
       || (n1==43 && (n2==41||n2==33))
       || (n1==42 && n2==28)
       || (n1==41 && (n2==40||n2==27||n2==22))
       || (n1==40 && (n2==21||n2==11))
       || (n1==39 && (n2==38||n2==36||n2==29||n2==22))
       || (n1==38 && (n2==25||n2==12))
       || (n1==37 && (n2==34||n2==24))
       || (n1==36 && (n2==23||n2==11))
       || (n1==34 && n2==15)
       || (n1==33 && n2==27)
       || (n1==29 && (n2==25||n2==23||n2==5))
       || (n1==28 && n2==18)
       || (n1==27 && (n2==20||n2==19||n2==17))
       || (n1==25 && n2==4)
       || (n1==24 && n2==15)
       || (n1==23 && (n2==16||n2==3))
       || (n1==22 && (n2==12||n2==11||n2==5))
       || (n1==21 && (n2==10||n2==9))
       || (n1==17 && n2==5)
       || (n1==16 && n2==2)
       || (n1==14 && n2==1)
       || (n1==12 && n2==4)
       || (n1==11 && n2==3)
       || (n1==10 && n2==4)
       || (n1== 9 && n2==3)
       || (n1== 6 && n2==1)
       || (n1== 5 && n2==3)) f1=0;
      s0=data10(n1,n2,&tau); break;
    case 11:
      s0=data11(n1,&tau); break;
  }
  if (f1) a=lireresolv(n1,n2,N,&nbv,&nbm);
  else { a = NULL; nbm=n1; nbv=n2; }

  ss = lirecoset(n1,n2,N); nbcoset = getcardgroup(ss);
  av1=avma; cpr=pr; nbgr = getcardgroup(tau);
  for (nogr=1; nogr<=nbgr; nogr++)
  {
    if (DEBUGLEVEL)
      { fprintferr("    ----> Group # %ld/%ld:\n",nogr,nbgr); flusherr(); }
    d = 1; f2 = f3 = 0;
    for(;;)
    {
      rr=r[d-1];
      if (DEBUGLEVEL && d>1)
        { fprintferr("       Working with polynomial #%ld:\n",d); flusherr(); }
      if (!f2)
      {
        f2=1;
        for(;;)
        {
          av2=avma; nbracmul=0; nbracint=0;
          for (nocoset=1; nocoset<=nbcoset; nocoset++)
          {
            gpol = get_pol_perm(tau[nogr], ss[nocoset], rr,a,nbm,nbv);
            sp = suffprec(gpol);
            if (sp>0) { avma=av1; break; }
            gpolint = is_int_pol(gpol,pr);
            if (gpolint)
            {
              nbracmul++;
              if (nbracmul>2520)
              {
                fprintferr("         more than 2520 rational integer roots\n");
                flusherr(); avma=av1; f2=0; break;
              }
              numerj[nbracmul]=nocoset; f1=1;
              for (j=1; j<=nbracint; j++)
                if (gegal(gpolint,racint[j]))
                  { multi[j]++; f1=0; norac[nbracmul]=j; break; }
              if (f1)
              {
                nbracint++;
                multi[nbracint]=1; numeri[nbracint]=nocoset;
                norac[nbracmul]=nbracint; tetpil=avma;
                racint[nbracint]=gerepile(av2,tetpil,gcopy(gpolint));
                av2=avma;
              }
            }
            avma=av2;
          }
          if (sp <= 0) break;

          cpr+=sp; nbdig=(cpr-2)*pariK;
          if (DEBUGLEVEL)
            { fprintferr("$$$$$ New precision = %ld\n",nbdig); flusherr(); }
          if (cpr<=PRMAX) preci(r,cpr);
          else 
          { 
            moreprec(po,r,pr,cpr-PRMAX); preci(r,cpr);
            av1=avma; rr=r[d-1];
          }
        }
        if (f2)
        {
          avma=av1;
          if (DEBUGLEVEL) dbg_rac(0,nbracint,numeri,racint,multi);
          if (nbracint)
          {
            f1=1;
            for (i=1; i<=nbracint; i++)
              if (multi[i]==1) { f1=0; uu=ss[numeri[i]]; break; }
            if (!f1) f3=1;
            else
            {
              lastnbri=nbracint; lastnbrm=nbracmul;
              for (j=1; j<=nbracmul; j++)
                { lastnum[j]=numerj[j]; lastnor[j]=norac[j]; }
            }
          }
          else f3=-1;
        }
      }
      else
      {
        nrm=0; nri=0;
        for (i=1; i<=lastnbri; i++)
        {
          for(;;)
          {
            av2=avma; nbracmul=nrm; nbracint=nri;
            for (j=1; j<=lastnbrm; j++)
              if (lastnor[j]==i)
              {
                gpol = get_pol_perm(tau[nogr], ss[lastnum[j]], rr,a,nbm,nbv);
                sp = suffprec(gpol);
                if (sp>0) { avma=av1; break; }
                gpolint = is_int_pol(gpol,pr);
                if (gpolint)
                {
                  nbracmul++;
                  numerj[nbracmul]=lastnum[j]; f1=1;
                  for (k=nri+1; k<=nbracint; k++)
                    if (gegal(gpolint,racint[k]))
                      { multi[k]++; f1=0; norac[nbracmul]=k; break; }
                  if (f1)
                  {
                    nbracint++;
                    multi[nbracint]=1; numeri[nbracint]=lastnum[j];
                    norac[nbracmul]=nbracint; tetpil=avma;
                    racint[nbracint]=gerepile(av2,tetpil,gcopy(gpolint));
                    av2=avma;
                  }
                }
                avma=av2;
              }
            if (sp <= 0) break; 

            cpr+=sp; nbdig=(cpr-2)*pariK;
            if (DEBUGLEVEL)
              { fprintferr("$$$$$ New precision = %ld\n",nbdig); flusherr(); }
            if (cpr<=PRMAX) preci(r,cpr);
            else
            { 
              moreprec(po,r,pr,cpr-PRMAX); preci(r,cpr);
              av1=avma; rr=r[d-1];
            }
          }
          avma=av1;
          if (DEBUGLEVEL) dbg_rac(nri,nbracint,numeri,racint,multi);
          if (nbracint>nri)
          {
            f1=1;
            for (k=nri+1; k<=nbracint; k++)
              if (multi[k]==1) { f1=0; uu=ss[numeri[k]]; break; }
            if (!f1) { f3=1; break; }
          }
          nri=nbracint; nrm=nbracmul;
        }
        if (!f3) 
          if (nbracint)
          {
            lastnbri=nbracint; lastnbrm=nbracmul;
            for (j=1; j<=nbracmul; j++)
              { lastnum[j]=numerj[j]; lastnor[j]=norac[j]; }
          }
          else f3=-1;
      }
      avma=av1; if (f3) break;
      if (DEBUGLEVEL)
      {
        fprintferr("         all rational integer roots are repeated roots\n");
        flusherr();
      }
      d++; if (d>TSCHMAX) { tschirn(po,r,cpr); av1=avma; }
    }
    if (f3==1)
    {
      vv=permmul(tau[nogr],uu); ww=permmul(vv,s0);
      if (DEBUGLEVEL)
      {
        fprintferr("         testing roots reordering: "); flusherr();
        gpol = get_pol(rr,uu,a,nbm,nbv);
        bruterr(gpol,'g',-1);
        fprintferr("\n    Output of isin_%ld_G_H(%ld,%ld): %ld .\n",N,n1,n2,n2);
        fprintferr("    Reordering of the roots : "); printperm(ww);
        fprintferr("\n"); flusherr();
      }
      avma=av1; rearrange(r,ww); free(vv); free(ww); break;
    }
  }
  desallocgroup(ss); desallocgroup(tau);
  if (a) desallocresolv(a,nbm);
  if (s0 != SID) free(s0);
  preci(r,pr);
  if (f3==1) return n2;
  if (DEBUGLEVEL)
  {
    fprintferr("    Output of isin_%ld_G_H(%ld,%ld): not included.\n",N,n1,n2);
    flusherr();
  }
  return 0;
}

/* BIBLIOTHEQUE POUR LE DEGRE 8 */

long
closure8(GEN po, long ca, long pr)
{
  long nbrac,rep1,rep2,rep;
  GEN p1,r[NMAX];

  r[0]=p1=roots(po,PRMAX); nbrac=lg(p1)-1;
  if (nbrac!=N) err(talker,"incompatible number of roots in closure8()");
  preci(r,pr);
  if (!ca)
  {
  /* CLOS_8_1: */
    rep1=isin_G_H(po,r,50,47,pr);
    if (!rep1)
    {
  /* CLOS_8_2: */
      rep2=isin_G_H(po,r,50,44,pr);
      if (!rep2) rep= galoisprim8(po,r,ca,pr);
      else rep=galoisimpodd8(po,r,pr,44);
    }
    else rep=galoisimpodd8(po,r,pr,47);
  }
  else
  {
  /* CLOS_8_3: */
    rep1=isin_G_H(po,r,49,45,pr);
    if (!rep1)
    {
    /* CLOS_8_4: */
      rep2=isin_G_H(po,r,49,39,pr);
      if (!rep2) rep=galoisprim8(po,r,ca,pr);
      else rep=galoisimpeven8(po,r,pr,39);
    }
    else rep = galoisimpeven8(po,r,pr,45);
  }
  return rep;
}

long
galoisprim8(GEN po, GEN *r, long ca, long pr)
{
  long rep;

/* PRIM_8_1: */
  rep=isin_G_H(po,r,50,43,pr);
  if (rep) return ca? 37: 43; 
/* PRIM_8_2: */
  if (!ca) return 50;
/* PRIM_8_3: */
  rep=isin_G_H(po,r,49,48,pr);
  if (!rep) return 49;
/* PRIM_8_4: */
  rep=isin_G_H(po,r,48,36,pr);
  if (!rep) return 48;
/* PRIM_8_5: */
  rep=isin_G_H(po,r,36,25,pr);
  return rep? 25: 36; 
}

long
galoisimpodd8(GEN po, GEN *r, long pr, long nh)
{
  long rep;
/* IMPODD_8_1: */
  if (nh!=47) goto IMPODD_8_6;
/* IMPODD_8_2: */
  rep=isin_G_H(po,r,47,46,pr);
  if (!rep) goto IMPODD_8_5;
/* IMPODD_8_4: */
  rep=isin_G_H(po,r,46,28,pr);
  if (rep) goto IMPODD_8_7; else return 46;

IMPODD_8_5:
  rep=isin_G_H(po,r,47,35,pr);
  if (rep) goto IMPODD_8_9; else return 47;

IMPODD_8_6:
  rep=isin_G_H(po,r,44,40,pr);
  if (rep) goto IMPODD_8_10; else goto IMPODD_8_11;

IMPODD_8_7:
  rep=isin_G_H(po,r,28,21,pr);
  if (rep) return 21; else goto IMPODD_8_33;

IMPODD_8_9:
  rep=isin_G_H(po,r,35,31,pr);
  if (rep) goto IMPODD_8_13; else goto IMPODD_8_14;

IMPODD_8_10:
  rep=isin_G_H(po,r,40,26,pr);
  if (rep) goto IMPODD_8_15; else goto IMPODD_8_16;

IMPODD_8_11:
  rep=isin_G_H(po,r,44,38,pr);
  if (rep) goto IMPODD_8_17; else goto IMPODD_8_18;

IMPODD_8_12:
  rep=isin_G_H(po,r,16,7,pr);
  if (rep) goto IMPODD_8_19; else return 16;

IMPODD_8_13:
  rep=isin_G_H(po,r,31,21,pr);
  return rep? 21: 31;

IMPODD_8_14:
  rep=isin_G_H(po,r,35,30,pr);
  if (rep) goto IMPODD_8_34; else goto IMPODD_8_20;

IMPODD_8_15:
  rep=isin_G_H(po,r,26,16,pr);
  if (rep) goto IMPODD_8_12; else goto IMPODD_8_21;

IMPODD_8_16:
  rep=isin_G_H(po,r,40,23,pr);
  if (rep) goto IMPODD_8_22; else return 40;

IMPODD_8_17:
  rep=isin_G_H(po,r,38,31,pr);
  if (rep) goto IMPODD_8_13; else return 38;

IMPODD_8_18:
  rep=isin_G_H(po,r,44,35,pr);
  if (rep) goto IMPODD_8_9; else return 44;

IMPODD_8_19:
  rep=isin_G_H(po,r,7,1,pr);
  return rep? 1: 7;

IMPODD_8_20:
  rep=isin_G_H(po,r,35,28,pr);
  if (rep) goto IMPODD_8_7; else goto IMPODD_8_23;

IMPODD_8_21:
  rep=isin_G_H(po,r,26,17,pr);
  if (rep) goto IMPODD_8_24; else goto IMPODD_8_25;

IMPODD_8_22:
  rep=isin_G_H(po,r,23,8,pr);
  if (rep) goto IMPODD_8_26; else return 23;

IMPODD_8_23:
  rep=isin_G_H(po,r,35,27,pr);
  if (rep) goto IMPODD_8_27; else goto IMPODD_8_28;

IMPODD_8_24:
  rep=isin_G_H(po,r,17,7,pr);
  if (rep) goto IMPODD_8_19; else return 17;

IMPODD_8_25:
  rep=isin_G_H(po,r,26,15,pr);
  if (rep) goto IMPODD_8_29; else return 26;

IMPODD_8_26:
  rep=isin_G_H(po,r,8,1,pr);
  return rep? 1: 8;

IMPODD_8_27:
  rep=isin_G_H(po,r,27,16,pr);
  if (rep) goto IMPODD_8_12; else return 27;

IMPODD_8_28:
  rep=isin_G_H(po,r,35,26,pr);
  if (rep) goto IMPODD_8_15; else return 35;

IMPODD_8_29:
  rep=isin_G_H(po,r,15,7,pr);
  if (rep) goto IMPODD_8_19;
/* IMPODD_8_30: */
  rep=isin_G_H(po,r,15,6,pr);
  if (!rep) goto IMPODD_8_32;
/* IMPODD_8_31: */
  rep=isin_G_H(po,r,6,1,pr);
  return rep? 1: 6;

IMPODD_8_32:
  rep=isin_G_H(po,r,15,8,pr);
  if (rep) goto IMPODD_8_26; else return 15;

IMPODD_8_33:
  rep=isin_G_H(po,r,28,16,pr);
  if (rep) goto IMPODD_8_12; else return 28;

IMPODD_8_34:
  rep=isin_G_H(po,r,30,21,pr);
  return rep? 21: 30;
}

long
galoisimpeven8(GEN po, GEN *r, long pr, long nh)
{
   long rep;
/* IMPEVEN_8_1: */
   if (nh!=45) goto IMPEVEN_8_6;
/* IMPEVEN_8_2: */
   rep=isin_G_H(po,r,45,42,pr);
   if (!rep) goto IMPEVEN_8_5;
/* IMPEVEN_8_4: */
  rep=isin_G_H(po,r,42,34,pr);
  if (rep) goto IMPEVEN_8_7; else goto IMPEVEN_8_8;

IMPEVEN_8_5:
  rep=isin_G_H(po,r,45,41,pr);
  if (rep) goto IMPEVEN_8_9; else return 45;

IMPEVEN_8_6:
  rep=isin_G_H(po,r,39,32,pr);
  if (rep) goto IMPEVEN_8_10; else goto IMPEVEN_8_11;

IMPEVEN_8_7:
  rep=isin_G_H(po,r,34,18,pr);
  if (rep) goto IMPEVEN_8_21; else goto IMPEVEN_8_45;

IMPEVEN_8_8:
  rep=isin_G_H(po,r,42,33,pr);
  if (rep) goto IMPEVEN_8_14; else return 42;

IMPEVEN_8_9:
  rep=isin_G_H(po,r,41,34,pr);
  if (rep) goto IMPEVEN_8_7; else goto IMPEVEN_8_15;

IMPEVEN_8_10:
  rep=isin_G_H(po,r,32,22,pr);
  if (rep) goto IMPEVEN_8_16; else goto IMPEVEN_8_17;

IMPEVEN_8_11:
  rep=isin_G_H(po,r,39,29,pr);
  if (rep) goto IMPEVEN_8_18; else goto IMPEVEN_8_19;

IMPEVEN_8_12:
  rep=isin_G_H(po,r,14,4,pr);
  return rep? 4: 14;

IMPEVEN_8_14:
  rep=isin_G_H(po,r,33,18,pr);
  if (rep) goto IMPEVEN_8_21; else goto IMPEVEN_8_22;

IMPEVEN_8_15:
  rep=isin_G_H(po,r,41,33,pr);
  if (rep) goto IMPEVEN_8_14; else goto IMPEVEN_8_23;

IMPEVEN_8_16:
  rep=isin_G_H(po,r,22,11,pr);
  if (rep) goto IMPEVEN_8_24; else goto IMPEVEN_8_25;

IMPEVEN_8_17:
  rep=isin_G_H(po,r,32,13,pr);
  if (rep) goto IMPEVEN_8_26; else goto IMPEVEN_8_27;

IMPEVEN_8_18:
  rep=isin_G_H(po,r,29,22,pr);
  if (rep) goto IMPEVEN_8_16; else goto IMPEVEN_8_28;

IMPEVEN_8_19:
  rep=isin_G_H(po,r,39,24,pr);
  if (rep) goto IMPEVEN_8_29; else return 39;

IMPEVEN_8_20:
  rep=isin_G_H(po,r,9,4,pr);
  if (rep) return 4; else goto IMPEVEN_8_30;

IMPEVEN_8_21:
  rep=isin_G_H(po,r,18,10,pr);
  if (rep) goto IMPEVEN_8_31; else goto IMPEVEN_8_32;

IMPEVEN_8_22:
  rep=isin_G_H(po,r,33,13,pr);
  if (rep) goto IMPEVEN_8_26; else return 33;

IMPEVEN_8_23:
  rep=isin_G_H(po,r,41,29,pr);
  if (rep) goto IMPEVEN_8_18; else goto IMPEVEN_8_33;

IMPEVEN_8_24:
  rep=isin_G_H(po,r,11,5,pr);
  if (rep) return 5; else goto IMPEVEN_8_34;

IMPEVEN_8_25:
  rep=isin_G_H(po,r,22,9,pr);
  if (rep) goto IMPEVEN_8_20; else return 22;

IMPEVEN_8_26:
  rep=isin_G_H(po,r,13,3,pr);
  return rep? 3: 13;

IMPEVEN_8_27:
  rep=isin_G_H(po,r,32,12,pr);
  if (rep) goto IMPEVEN_8_35; else return 32;

IMPEVEN_8_28:
  rep=isin_G_H(po,r,29,20,pr);
  if (rep) goto IMPEVEN_8_36; else goto IMPEVEN_8_37;

IMPEVEN_8_29:
  rep=isin_G_H(po,r,24,14,pr);
  if (rep) goto IMPEVEN_8_12; else goto IMPEVEN_8_38;

IMPEVEN_8_30:
  rep=isin_G_H(po,r,9,3,pr);
  if (rep) return 3; else goto IMPEVEN_8_39;

IMPEVEN_8_31:
  rep=isin_G_H(po,r,10,2,pr);
  return rep? 2: 10;

IMPEVEN_8_32:
  rep=isin_G_H(po,r,18,9,pr);
  if (rep) goto IMPEVEN_8_20; else return 18;

IMPEVEN_8_33:
  rep=isin_G_H(po,r,41,24,pr);
  if (rep) goto IMPEVEN_8_29; else return 41;

IMPEVEN_8_34:
  rep=isin_G_H(po,r,11,4,pr);
  if (rep) return 4; else goto IMPEVEN_8_44;

IMPEVEN_8_35:
  rep=isin_G_H(po,r,12,5,pr);
  return rep? 5: 12;

IMPEVEN_8_36:
  rep=isin_G_H(po,r,20,10,pr);
  if (rep) goto IMPEVEN_8_31; else return 20;

IMPEVEN_8_37:
  rep=isin_G_H(po,r,29,19,pr);
  if (rep) goto IMPEVEN_8_40; else goto IMPEVEN_8_41;

IMPEVEN_8_38:
  rep=isin_G_H(po,r,24,13,pr);
  if (rep) goto IMPEVEN_8_26; else goto IMPEVEN_8_42;

IMPEVEN_8_39:
  rep=isin_G_H(po,r,9,2,pr);
  return rep? 2: 9;

IMPEVEN_8_40:
  rep=isin_G_H(po,r,19,10,pr);
  if (rep) goto IMPEVEN_8_31; else goto IMPEVEN_8_43;

IMPEVEN_8_41:
  rep=isin_G_H(po,r,29,18,pr);
  if (rep) goto IMPEVEN_8_21; else return 29;

IMPEVEN_8_42:
  rep=isin_G_H(po,r,24,9,pr);
  if (rep) goto IMPEVEN_8_20; else return 24;

IMPEVEN_8_43:
  rep=isin_G_H(po,r,19,9,pr);
  if (rep) goto IMPEVEN_8_20; else return 19;

IMPEVEN_8_44:
  rep=isin_G_H(po,r,11,2,pr);
  return rep? 2: 11;

IMPEVEN_8_45:
  rep=isin_G_H(po,r,34,14,pr);
  if (rep) goto IMPEVEN_8_12; else return 34;
}

static GROUP
initgroup(long n, long nbgr)
{
  GROUP t = allocgroup(n,nbgr);
  t[1] = SID; return t;
}

static PERMUTATION
data8(long n1, long n2, GROUP *t)
{
  SID[0] = N;
  switch(n1)
  {
    case 7: if (n2!=1) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 6, 5, 8, 7);
      return SID;
    case 9: if (n2!=4) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 4, 3, 5, 6, 8, 7);
      return SID;
    case 10: if (n2!=2) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 6, 5, 8, 7);
      return SID;
    case 11:
      switch(n2)
      {
        case 2: 
          *t=initgroup(N,2);
          _aff((*t)[2], 1, 2, 5, 6, 3, 4, 8, 7);
          return _cr(1, 3, 5, 8, 2, 4, 6, 7);
        case 4:      
          *t=initgroup(N,1);
          return _cr(1, 3, 7, 5, 2, 4, 8, 6);
      }break;
    case 14: if (n2!=4) break;
      *t=initgroup(N,1);
      return _cr(1, 2, 4, 3, 5, 6, 8, 7); 
    case 15: if (n2!=6 && n2!=8) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 6, 5, 8, 7);
      return SID;
    case 16: if (n2!=7) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 7);
      return SID;
    case 18:
      switch(n2)
      {
        case 9: *t=initgroup(N,3);
          _aff((*t)[2], 1, 5, 3, 7, 2, 6, 4, 8);
          _aff((*t)[3], 1, 2, 3, 4, 6, 5, 8, 7);
          return SID;
        case 10: *t=initgroup(N,3);
          _aff((*t)[2], 1, 6, 3, 8, 2, 5, 4, 7);
          _aff((*t)[3], 1, 5, 3, 7, 2, 6, 4, 8);
          return SID;
      }break;
    case 19: if (n2!=9) break;
      *t=initgroup(N,1);
      return _cr(1, 5, 3, 8, 2, 6, 4, 7); 
    case 20: if (n2!=10) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 7);
      return SID;
    case 22:
      switch(n2)
      {
        case 9: *t=initgroup(N,6);
          _aff((*t)[2], 1, 2, 7, 8, 3, 4, 6, 5);
          _aff((*t)[3], 1, 2, 7, 8, 3, 4, 5, 6);
          _aff((*t)[4], 1, 2, 5, 6, 3, 4, 8, 7);
          _aff((*t)[5], 1, 2, 5, 6, 3, 4, 7, 8);
          _aff((*t)[6], 1, 2, 3, 4, 5, 6, 8, 7);
          return _cr(1, 3, 5, 7, 2, 4, 6, 8);
        case 11: *t=initgroup(N,6);
          _aff((*t)[2], 1, 2, 5, 6, 7, 8, 4, 3);
          _aff((*t)[3], 1, 2, 5, 6, 7, 8, 3, 4);
          _aff((*t)[4], 1, 2, 3, 4, 7, 8, 6, 5);
          _aff((*t)[5], 1, 2, 3, 4, 7, 8, 5, 6);
          _aff((*t)[6], 1, 2, 3, 4, 5, 6, 8, 7);
          return SID;
      }break;
    case 23: if (n2!=8) break;
      *t=initgroup(N,1); 
      return _cr(1, 2, 3, 4, 6, 5, 8, 7); 
    case 26: if (n2!=15 && n2!=17) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 7);
      return SID;
    case 28: if (n2!=21) break;
      *t=initgroup(N,1); 
      return _cr(1, 2, 3, 4, 7, 8, 5, 6); 
    case 29: if (n2!=18 && n2!=19) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 7);
      return SID;
    case 30: if (n2!=21) break;
      *t=initgroup(N,1);
      return _cr(1, 2, 3, 4, 7, 8, 5, 6); 
    case 31: if (n2!=21) break;
      *t=initgroup(N,3);
      _aff((*t)[2], 1, 2, 3, 4, 7, 8, 5, 6);
      _aff((*t)[3], 1, 2, 5, 6, 7, 8, 3, 4);
      return SID;
    case 32: if (n2!=12 && n2!=13) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 7);
      return SID;
    case 33:
      switch(n2)
      {
        case 13: *t=initgroup(N,1); 
          return _cr(1, 5, 2, 6, 3, 7, 4, 8); 
        case 18: *t=initgroup(N,1); 
          return _cr(1, 2, 5, 6, 3, 4, 7, 8); 
      }break;
    case 34:
      switch(n2)
      {
        case 14: *t=initgroup(N,3);
          _aff((*t)[2], 1, 2, 3, 4, 5, 8, 6, 7);
          _aff((*t)[3], 1, 2, 3, 4, 5, 7, 8, 6);
          return _cr(1, 5, 2, 6, 3, 7, 4, 8);
        case 18: *t=initgroup(N,1);
          return _cr(1, 2, 5, 6, 3, 4, 8, 7); 
      }break;
    case 39: if (n2!=24) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 7);
      return SID;
    case 40: if (n2!=23) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 7);
      return SID;
    case 41:
      switch(n2)
      {
        case 24: *t=initgroup(N,1); 
          return _cr(1, 5, 2, 6, 3, 7, 4, 8); 
        case 29: *t=initgroup(N,1); 
          return _cr(1, 2, 5, 6, 3, 4, 7, 8); 
      }break;
    case 42: if (n2!=34) break;
      *t=initgroup(N,1);
      return _cr(1, 2, 3, 4, 5, 6, 8, 7); 
    case 45: if (n2!=41 && n2!=42) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 7);
      return SID;
    case 46: if (n2!=28) break;
      *t=initgroup(N,1);
      return _cr(1, 2, 5, 6, 3, 4, 7, 8); 
    case 47: if (n2!=35) break;
      *t=initgroup(N,1);
      return _cr(1, 2, 5, 6, 3, 4, 7, 8); 
    case 49: if (n2!=48) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 7);
      return SID;
  }
  *t=initgroup(N,1); return SID;
}

long
galoismodulo8(GEN pol, GEN dpol, long car)
{
  byteptr d=diffptr;
  GEN ff,p1;
  long av = avma, i,j,k,p,nbgroup,numer,nbraye,av1,gr[51];
  long lp1,galtyp[NMAX+1],nbremain;
  long **TYP8 = partitions(8);
  long **GR8 = (long**)cgeti(49);

/* List of possible types in group j: GR[j][0] = #GR[j] if
 * the group is odd, - #GR[j] if even */
  GR8[ 1]= _affl(  4, 1,5,15,22);
  GR8[ 2]= _affl( -3, 1,5,15);
  GR8[ 3]= _affl( -2, 1,5);
  GR8[ 4]= _affl( -3, 1,5,15);
  GR8[ 5]= _affl( -3, 1,5,15);
  GR8[ 6]= _affl(  5, 1,4,5,15,22);
  GR8[ 7]= _affl(  5, 1,3,5,15,22);
  GR8[ 8]= _affl(  5, 1,4,5,15,22);
  GR8[ 9]= _affl( -4, 1,3,5,15);
  GR8[10]= _affl( -4, 1,3,5,15);
  GR8[11]= _affl( -4, 1,3,5,15);
  GR8[12]= _affl( -5, 1,5,9,15,20);
  GR8[13]= _affl( -4, 1,5,9,20);
  GR8[14]= _affl( -4, 1,5,9,15);
  GR8[15]= _affl(  6, 1,3,4,5,15,22);
  GR8[16]= _affl(  5, 1,3,5,15,22);
  GR8[17]= _affl(  7, 1,3,5,11,13,15,22);
  GR8[18]= _affl( -4, 1,3,5,15);
  GR8[19]= _affl( -5, 1,3,5,12,15);
  GR8[20]= _affl( -4, 1,3,5,15);
  GR8[21]= _affl(  5, 1,3,5,13,15);
  GR8[22]= _affl( -4, 1,3,5,15);
  GR8[23]= _affl(  7, 1,4,5,9,15,20,22);
  GR8[24]= _affl( -6, 1,3,5,9,15,20);
  GR8[25]= _affl( -3, 1,5,21);
  GR8[26]= _affl(  8, 1,3,4,5,11,13,15,22);
  GR8[27]= _affl(  8, 1,2,3,4,5,13,15,22);
  GR8[28]= _affl(  7, 1,3,5,12,13,15,22);
  GR8[29]= _affl( -5, 1,3,5,12,15);
  GR8[30]= _affl(  7, 1,3,4,5,11,13,15);
  GR8[31]= _affl(  7, 1,2,3,4,5,13,15);
  GR8[32]= _affl( -6, 1,3,5,9,15,20);
  GR8[33]= _affl( -6, 1,3,5,9,15,20);
  GR8[34]= _affl( -5, 1,3,5,9,15);
  GR8[35]= _affl( 10, 1,2,3,4,5,11,12,13,15,22);
  GR8[36]= _affl( -5, 1,5,9,20,21);
  GR8[37]= _affl( -5, 1,5,9,15,21);
  GR8[38]= _affl( 11, 1,2,3,4,5,9,10,13,15,19,20);
  GR8[39]= _affl( -7, 1,3,5,9,12,15,20);
  GR8[40]= _affl( 10, 1,3,4,5,9,11,13,15,20,22);
  GR8[41]= _affl( -7, 1,3,5,9,12,15,20);
  GR8[42]= _affl( -8, 1,3,5,6,8,9,15,20);
  GR8[43]= _affl(  8, 1,4,5,9,15,19,21,22);
  GR8[44]= _affl( 14, 1,2,3,4,5,9,10,11,12,13,15,19,20,22);
  GR8[45]= _affl( -9, 1,3,5,6,8,9,12,15,20);
  GR8[46]= _affl( 10, 1,3,5,6,8,9,12,13,15,22);
  GR8[47]= _affl( 16, 1,2,3,4,5,6,7,8,9,11,12,13,14,15,20,22);
  GR8[48]= _affl( -8, 1,3,5,9,12,15,20,21);

  nbgroup=49; nbremain = car? 28: 22;
  p = 0; for (i=1; i<51; i++) gr[i]=1;
  for (k=1; k<15; k++, d++)
  {
    p += *d; if (!*d) err(primer1);
    if (smodis(dpol,p)) /* p does not divide dpol */
    {
      av1=avma; ff=simplefactmod(pol,stoi(p));
      p1=(GEN)ff[1]; lp1=lg(p1);
      for (j=1; j<lp1; j++) galtyp[j]=itos((GEN) p1[lp1-j]);
      for (   ; j<=N;  j++) galtyp[j]=0;
      ranger(galtyp,N); numer=numerotyp(TYP8,galtyp);
      if (car) nbraye = rayergrouppair(GR8,numer,gr,nbgroup);
      else     nbraye = rayergroupimpair(GR8,numer,gr,nbgroup);
      nbremain -= nbraye;
      if (nbremain==1) { avma=av; return car? 49: 50; }
      avma=av1; 
    }
  }
  avma=av; return 0;
}

/* BIBLIOTHEQUE POUR LE DEGRE 9 */

long
closure9(GEN po, long ca, long pr)
{
  long nbrac,rep;
  GEN p1,r[NMAX];

  r[0]=p1=roots(po,PRMAX); nbrac=lg(p1)-1;
  if (nbrac!=N) err(talker,"incompatible number of roots in closure9()");
  preci(r,pr);
  if (!ca)
  {
  /* CLOS_9_1: */
    rep=isin_G_H(po,r,34,31,pr);
    if (!rep) rep=galoisprim9(po,r,ca,pr);
    else rep=galoisimpodd9(po,r,pr);
  }
  else
  {
  /* CLOS_9_2: */
    rep=isin_G_H(po,r,33,30,pr);
    if (!rep) rep=galoisprim9(po,r,ca,pr);
    else rep=galoisimpeven9(po,r,pr);
  }
  return rep;
}

long
galoisprim9(GEN po, GEN *r, long ca, long pr)
{
  long rep;

  if (!ca)
  {
  /* PRIM_9_1: */
    rep=isin_G_H(po,r,34,26,pr);
    if (!rep) return 34;
  /* PRIM_9_2: */
    rep=isin_G_H(po,r,26,19,pr);
    if (!rep) return 26;
  /* PRIM_9_3: */
    rep=isin_G_H(po,r,19,16,pr);
    if (rep) return 16;
  /* PRIM_9_4: */
    rep=isin_G_H(po,r,19,15,pr);
    return rep? 15: 19;
  }
/* PRIM_9_5: */
  rep=isin_G_H(po,r,33,32,pr);
  if (!rep) goto PRIM_9_7;
/* PRIM_9_6: */
  rep=isin_G_H(po,r,32,27,pr);
  return rep? 27: 32;

PRIM_9_7:
  rep=isin_G_H(po,r,33,23,pr);
  if (!rep) return 33;
/* PRIM_9_8: */
  rep=isin_G_H(po,r,23,14,pr);
  if (!rep) return 23;
/* PRIM_9_9: */
  rep=isin_G_H(po,r,14,9,pr);
  return rep? 9: 14;
}

long
galoisimpodd9(GEN po, GEN *r, long pr)
{
  long rep;

/* IMPODD_9_1: */
  rep=isin_G_H(po,r,31,29,pr);
  if (!rep) goto IMPODD_9_5;
/* IMPODD_9_2: */
  rep=isin_G_H(po,r,29,20,pr);
  if (!rep) return 29;
IMPODD_9_3:
  rep=isin_G_H(po,r,20,12,pr);
  if (!rep) return 20;
IMPODD_9_4:
  rep=isin_G_H(po,r,12,4,pr);
  return rep? 4: 12;

IMPODD_9_5:
  rep=isin_G_H(po,r,31,28,pr);
  if (!rep) goto IMPODD_9_9;
/* IMPODD_9_6: */
  rep=isin_G_H(po,r,28,22,pr);
  if (!rep) return 28;
IMPODD_9_7:
  rep=isin_G_H(po,r,22,13,pr);
  if (!rep) return 22;
IMPODD_9_8:
  rep=isin_G_H(po,r,13,4,pr);
  return rep? 4: 13;

IMPODD_9_9:
  rep=isin_G_H(po,r,31,24,pr);
  if (!rep) return 31;
/* IMPODD_9_10: */
  rep=isin_G_H(po,r,24,22,pr);
  if (rep) goto IMPODD_9_7;
/* IMPODD_9_11: */
  rep=isin_G_H(po,r,24,20,pr);
  if (rep) goto IMPODD_9_3;
/* IMPODD_9_12: */
  rep=isin_G_H(po,r,24,18,pr);
  if (!rep) return 24;
/* IMPODD_9_13: */
  rep=isin_G_H(po,r,18,13,pr);
  if (rep) goto IMPODD_9_8;
/* IMPODD_9_14: */
  rep=isin_G_H(po,r,18,12,pr);
  if (rep) goto IMPODD_9_4;
/* IMPODD_9_15: */
  rep=isin_G_H(po,r,18,8,pr);
  if (!rep) return 18;
/* IMPODD_9_16: */
  rep=isin_G_H(po,r,8,4,pr);
  return rep? 4: 8;
}

long
galoisimpeven9(GEN po, GEN *r, long pr)
{
  long rep;

/* IMPEVEN_9_1: */
  rep=isin_G_H(po,r,30,25,pr);
  if (!rep) goto IMPEVEN_9_7;
/* IMPEVEN_9_2: */
  rep=isin_G_H(po,r,25,17,pr);
  if (!rep) return 25;
IMPEVEN_9_3:
  rep=isin_G_H(po,r,17,7,pr);
  if (!rep) goto IMPEVEN_9_5;
IMPEVEN_9_4:
  rep=isin_G_H(po,r,7,2,pr);
  return rep? 2: 7;

IMPEVEN_9_5:
  rep=isin_G_H(po,r,17,6,pr);
  if (!rep) return 17;
IMPEVEN_9_6:
  rep=isin_G_H(po,r,6,1,pr);
  return rep? 1: 6;

IMPEVEN_9_7:
  rep=isin_G_H(po,r,30,21,pr);
  if (!rep) return 30;
/* IMPEVEN_9_8: */
  rep=isin_G_H(po,r,21,17,pr);
  if (rep) goto IMPEVEN_9_3;
/* IMPEVEN_9_9: */
  rep=isin_G_H(po,r,21,11,pr);
  if (!rep) goto IMPEVEN_9_13;
/* IMPEVEN_9_10: */
  rep=isin_G_H(po,r,11,7,pr);
  if (rep) goto IMPEVEN_9_4;
/* IMPEVEN_9_11: */
  rep=isin_G_H(po,r,11,5,pr);
  if (!rep) return 11;
/* IMPEVEN_9_12: */
  rep=isin_G_H(po,r,5,2,pr);
  return rep? 2: 5;

IMPEVEN_9_13:
  rep=isin_G_H(po,r,21,10,pr);
  if (!rep) return 21;
/* IMPEVEN_9_14: */
  rep=isin_G_H(po,r,10,6,pr);
  if (rep) goto IMPEVEN_9_6;
/* IMPEVEN_9_15: */
  rep=isin_G_H(po,r,10,3,pr);
  if (!rep) return 10;
/* IMPEVEN_9_16: */
  rep=isin_G_H(po,r,3,1,pr);
  return rep? 1: 3;
}

PERMUTATION
data9(long n1, long n2, GROUP *t)
{
  SID[0]=N;
  switch(n1)
  {
    case 6: if (n2!=1) break;
      *t=initgroup(N,3);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 9, 7);
      _aff((*t)[3], 1, 2, 3, 4, 5, 6, 9, 7, 8);
      return SID;
    case 7: if (n2!=2) break;
      *t=initgroup(N,3);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 9, 7);
      _aff((*t)[3], 1, 2, 3, 4, 5, 6, 9, 7, 8);
      return SID;
    case 8: if (n2!=4) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 4, 7, 2, 5, 8, 3, 6, 9);
      return SID;
    case 12: if (n2!=4) break;
      *t=initgroup(N,3);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 9, 7);
      _aff((*t)[3], 1, 2, 3, 4, 5, 6, 9, 7, 8);
      return SID;
    case 13: if (n2!=4) break;
      *t=initgroup(N,1);
      return _cr(1, 4, 7, 2, 5, 8, 3, 6, 9); 
    case 14: if (n2!=9) break;
      *t=initgroup(N,3);
      _aff((*t)[2], 1, 2, 3, 5, 6, 4, 9, 7, 8);
      _aff((*t)[3], 1, 2, 3, 6, 4, 5, 8, 9, 7);
      return SID;
    case 17: if (n2!=6) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 7, 8, 9, 4, 5, 6);
      return SID;
    case 21: if (n2!=10) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 7, 8, 9, 4, 5, 6);
      return SID;
    case 33: if (n2!=32) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 7, 9, 8);
      return SID;
  }
  *t=initgroup(N,1); return SID;
}

long
galoismodulo9(GEN pol, GEN dpol, long car)
{
  byteptr d=diffptr;
  GEN ff,p1;
  long av = avma, av1,i,k,p,nbgroup,numer,nbraye,gr[35];
  long lp1,galtyp[NMAX+1],nbremain,j;
  long **TYP9 = partitions(9);
  long **GR9 = (long**) cgeti(33);

  /* 42 TYPES ORDONNES CROISSANT (T[1],...,T[30])*/

  GR9[ 1]= _affl( -3, 1,12,30);
  GR9[ 2]= _affl( -2, 1,12);
  GR9[ 3]= _affl( -4, 1,5,12,30);
  GR9[ 4]= _affl(  4, 1,4,12,26);
  GR9[ 5]= _affl( -3, 1,5,12);
  GR9[ 6]= _affl( -4, 1,10,12,30);
  GR9[ 7]= _affl( -3, 1,10,12);
  GR9[ 8]= _affl(  5, 1,4,5,12,26);
  GR9[ 9]= _affl( -4, 1,5,12,18);
  GR9[10]= _affl( -6, 1,5,10,12,25,30);
  GR9[11]= _affl( -5, 1,5,10,12,25);
  GR9[12]= _affl(  5, 1,4,10,12,26);
  GR9[13]= _affl(  5, 1,4,10,12,26);
  GR9[14]= _affl( -4, 1,5,12,18);
  GR9[15]= _affl(  5, 1,5,12,18,29);
  GR9[16]= _affl(  6, 1,4,5,12,18,26);
  GR9[17]= _affl( -5, 1,6,10,12,30);
  GR9[18]= _affl(  7, 1,4,5,10,12,25,26);
  GR9[19]= _affl(  7, 1,4,5,12,18,26,29);
  GR9[20]= _affl(  9, 1,4,6,9,10,12,24,26,30);
  GR9[21]= _affl( -7, 1,5,6,10,12,25,30);
  GR9[22]= _affl(  7, 1,4,6,10,12,26,30);
  GR9[23]= _affl( -6, 1,5,10,12,18,25);
  GR9[24]= _affl( 11, 1,4,5,6,9,10,12,24,25,26,30);
  GR9[25]= _affl( -7, 1,3,6,8,10,12,30);
  GR9[26]= _affl(  9, 1,4,5,10,12,18,25,26,29);
  GR9[27]= _affl( -5, 1,5,12,27,30);
  GR9[28]= _affl( 12, 1,2,3,4,6,7,8,10,11,12,26,30);
  GR9[29]= _affl( 12, 1,3,4,6,8,9,10,12,15,24,26,30);
  GR9[30]= _affl(-11, 1,3,5,6,8,10,12,14,17,25,30);
  GR9[31]= _affl( 19, 1,2,3,4,5,6,7,8,9,10,11,12,14,15,17,24,25,26,30);
  GR9[32]= _affl( -7, 1,5,10,12,25,27,30);

  nbgroup=33; nbremain = car? 18: 16;
  p = 0; for (i=1; i<35; i++) gr[i]=1;
  for (k=1; k<15; k++, d++)
  {
    p += *d; if (!*d) err(primer1);
    if (smodis(dpol,p)) /* p does not divide dpol */
    {
      av1=avma; ff=simplefactmod(pol,stoi(p)); 
      p1=(GEN)ff[1]; lp1=lg(p1);
      for (j=1; j<lp1; j++) galtyp[j]=itos((GEN)p1[lp1-j]);
      for (   ; j<=N ; j++) galtyp[j]=0;
      ranger(galtyp,N); numer=numerotyp(TYP9,galtyp);
      if (car) nbraye = rayergrouppair(GR9,numer,gr,nbgroup);
      else     nbraye = rayergroupimpair(GR9,numer,gr,nbgroup);
      nbremain -= nbraye;
      if (nbremain==1) { avma=av; return car? 33: 34; }
      avma=av1; k++; p=p+(*(++d));
    }
  }
  avma=av; return 0;
}

/* BIBLIOTHEQUE POUR LE DEGRE 10 */

long
closure10(GEN po, long ca, long pr)
{
  long nbrac,rep;
  GEN p1,r[NMAX];

  r[0]=p1=roots(po,PRMAX); nbrac=lg(p1)-1;
  if (nbrac!=N) err(talker,"incompatible number of roots in closure10()");

  preci(r,pr);
  if (ca)
  {
   /* CLOS_10_1: */
    rep=isin_G_H(po,r,44,42,pr);
    if (rep) rep=galoisimpeven10(po,r,pr,42);
    else
    {
   /* CLOS_10_2: */
     rep=isin_G_H(po,r,44,37,pr);
     if (rep) rep=galoisimpeven10(po,r,pr,37);
     else rep=galoisprim10(po,r,ca,pr);
    }
  }
  else
  {
   /* CLOS_10_3: */
    rep=isin_G_H(po,r,45,43,pr);
    if (rep) rep=galoisimpodd10(po,r,pr,43);
    else
    {
   /* CLOS_10_4: */
     rep=isin_G_H(po,r,45,39,pr);
     if (rep) rep=galoisimpodd10(po,r,pr,39);
     else rep=galoisprim10(po,r,ca,pr);
    }
  }
  return rep;
}

long
galoisprim10(GEN po, GEN *r, long ca, long pr)
{
  long rep;
  if (ca)
  {
  /* PRIM_10_1: */
    rep=isin_G_H(po,r,44,31,pr);
    if (!rep) return 44;
  /* PRIM_10_2: */
    rep=isin_G_H(po,r,31,26,pr);
    if (!rep) return 31;
  /* PRIM_10_3: */
    rep=isin_G_H(po,r,26,7,pr);
    return rep? 7: 26;
  }
  else
  {
  /* PRIM_10_4: */
    rep=isin_G_H(po,r,45,35,pr);
    if (!rep) return 45;
  /* PRIM_10_5: */
    rep=isin_G_H(po,r,35,32,pr);
    if (!rep) goto PRIM_10_7;
  /* PRIM_10_6: */
    rep=isin_G_H(po,r,32,13,pr);
    return rep? 13: 32;

   PRIM_10_7:
    rep=isin_G_H(po,r,35,30,pr);
    return rep? 30: 35;
  }
}

long
galoisimpeven10(GEN po, GEN *r, long pr, long nogr)
{
  long rep;
  if (nogr==42)
  {
 /* IMPEVEN_10_1: */
    rep=isin_G_H(po,r,42,28,pr);
    if (!rep) return 42;
 /* IMPEVEN_10_2: */
    rep=isin_G_H(po,r,28,18,pr);
    return rep? 18: 28;
  }
  else
  {
 /* IMPEVEN_10_3: */
    rep=isin_G_H(po,r,37,34,pr);
    if (!rep) goto IMPEVEN_10_5;
 /* IMPEVEN_10_4: */
    rep=isin_G_H(po,r,34,15,pr);
    if (rep) goto IMPEVEN_10_7; else return 34;

  IMPEVEN_10_5:
    rep=isin_G_H(po,r,37,24,pr);
    if (!rep) return 37;
 /* IMPEVEN_10_6: */
    rep=isin_G_H(po,r,24,15,pr);
    if (!rep) return 24;
  IMPEVEN_10_7:
    rep=isin_G_H(po,r,15,8,pr);
    return rep? 8: 15;
  }
}

long
galoisimpodd10(GEN po, GEN *r, long pr, long nogr)
{
  long rep;
  if (nogr==43)
  {
 /*  IMPODD_10_1: */
    rep=isin_G_H(po,r,43,41,pr);
    if (!rep) goto IMPODD_10_3;
 /* IMPODD_10_2: */
    rep=isin_G_H(po,r,41,40,pr);
    if (rep) goto IMPODD_10_4; else goto IMPODD_10_5;

   IMPODD_10_3:
    rep=isin_G_H(po,r,43,33,pr);
    if (rep) goto IMPODD_10_6; else return 43;

   IMPODD_10_4:
    rep=isin_G_H(po,r,40,21,pr);
    if (rep) goto IMPODD_10_7; else goto IMPODD_10_8;

   IMPODD_10_5:
    rep=isin_G_H(po,r,41,27,pr);
    if (rep) goto IMPODD_10_9; else goto IMPODD_10_10;

   IMPODD_10_6:
    rep=isin_G_H(po,r,33,27,pr);
    if (rep) goto IMPODD_10_9; else return 33;

   IMPODD_10_7:
    rep=isin_G_H(po,r,21,10,pr);
    if (rep) goto IMPODD_10_12; else goto IMPODD_10_13;

   IMPODD_10_8:
    rep=isin_G_H(po,r,40,12,pr);
    if (rep) goto IMPODD_10_14; else goto IMPODD_10_15;

   IMPODD_10_9:
    rep=isin_G_H(po,r,27,21,pr);
    if (rep) goto IMPODD_10_7; else goto IMPODD_10_16;

   IMPODD_10_10:
    rep=isin_G_H(po,r,41,22,pr);
    if (!rep) return 41;
 /* IMPODD_10_11: */
    rep=isin_G_H(po,r,22,12,pr);
    if (rep) goto IMPODD_10_14; else goto IMPODD_10_18;

   IMPODD_10_12:
    rep=isin_G_H(po,r,10,4,pr);
    return rep? 4: 10;

   IMPODD_10_13:
    rep=isin_G_H(po,r,21,9,pr);
    if (rep) goto IMPODD_10_19; else return 21;
   IMPODD_10_14:
    rep=isin_G_H(po,r,12,4,pr);
    return rep? 4: 12;

   IMPODD_10_15:
    rep=isin_G_H(po,r,40,11,pr);
    if (rep) goto IMPODD_10_20; else return 40;
   IMPODD_10_16:
    rep=isin_G_H(po,r,27,20,pr);
    if (!rep) goto IMPODD_10_21;
 /* IMPODD_10_17: */
    rep=isin_G_H(po,r,20,10,pr);
    if (rep) goto IMPODD_10_12; return 20;

   IMPODD_10_18:
    rep=isin_G_H(po,r,22,11,pr);
    if (rep) goto IMPODD_10_20; else goto IMPODD_10_23;

   IMPODD_10_19:
    rep=isin_G_H(po,r,9,6,pr);
    if (rep) goto IMPODD_10_24; else goto IMPODD_10_25;

   IMPODD_10_20:
    rep=isin_G_H(po,r,11,3,pr);
    if (rep) goto IMPODD_10_26; else return 11;

   IMPODD_10_21:
    rep=isin_G_H(po,r,27,19,pr);
    if (rep) goto IMPODD_10_27;
 /* IMPODD_10_22: */
    rep=isin_G_H(po,r,27,17,pr);
    if (rep) goto IMPODD_10_28; else return 27;

   IMPODD_10_23:
    rep=isin_G_H(po,r,22,5,pr);
    if (rep) goto IMPODD_10_29; else return 22;

   IMPODD_10_24:
    rep=isin_G_H(po,r,6,2,pr);
    if (rep) return 2; else goto IMPODD_10_30;

   IMPODD_10_25:
    rep=isin_G_H(po,r,9,3,pr);
    if (!rep) return 9;
   IMPODD_10_26:
    rep=isin_G_H(po,r,3,2,pr);
    if (rep) return 2; else goto IMPODD_10_31;

   IMPODD_10_27:
    rep=isin_G_H(po,r,19,9,pr);
    if (rep) goto IMPODD_10_19; else return 19;

   IMPODD_10_28:
    rep=isin_G_H(po,r,17,10,pr);
    if (rep) goto IMPODD_10_12; else goto IMPODD_10_32;

   IMPODD_10_29:
    rep=isin_G_H(po,r,5,4,pr);
    if (rep) return 4; else goto IMPODD_10_33;

   IMPODD_10_30:
    rep=isin_G_H(po,r,6,1,pr);
    return rep? 1: 6;

   IMPODD_10_31:
    rep=isin_G_H(po,r,3,1,pr);
    return rep? 1: 3;

   IMPODD_10_32:
    rep=isin_G_H(po,r,17,9,pr);
    if (rep) goto IMPODD_10_19; else goto IMPODD_10_60;

   IMPODD_10_33:
    rep=isin_G_H(po,r,5,3,pr);
    if (rep) goto IMPODD_10_26; else return 5;

   IMPODD_10_60:
    rep=isin_G_H(po,r,17,5,pr);
    if (rep) goto IMPODD_10_29; else return 17;
  }
  else
  {
  /* IMPODD_10_34: */
    rep=isin_G_H(po,r,39,38,pr);
    if (!rep) goto IMPODD_10_36;
  /* IMPODD_10_35: */
    rep=isin_G_H(po,r,38,25,pr);
    if (rep) goto IMPODD_10_37; else goto IMPODD_10_38;

   IMPODD_10_36:
    rep=isin_G_H(po,r,39,36,pr);
    if (rep) goto IMPODD_10_39; else goto IMPODD_10_40;

   IMPODD_10_37:
    rep=isin_G_H(po,r,25,4,pr);
    return rep? 4: 25;

   IMPODD_10_38:
    rep=isin_G_H(po,r,38,12,pr);
    if (rep) goto IMPODD_10_41; else return 38;

   IMPODD_10_39:
    rep=isin_G_H(po,r,36,23,pr);
    if (rep) goto IMPODD_10_42; else goto IMPODD_10_43;

   IMPODD_10_40:
    rep=isin_G_H(po,r,39,29,pr);
    if (rep) goto IMPODD_10_44; else goto IMPODD_10_45;

   IMPODD_10_41:
    rep=isin_G_H(po,r,12,4,pr);
    return rep? 4: 12;

   IMPODD_10_42:
    rep=isin_G_H(po,r,23,16,pr);
    if (rep) goto IMPODD_10_46; else goto IMPODD_10_47;

   IMPODD_10_43:
    rep=isin_G_H(po,r,36,11,pr);
    if (rep) goto IMPODD_10_48; else return 36;

   IMPODD_10_44:
    rep=isin_G_H(po,r,29,25,pr);
    if (rep) goto IMPODD_10_37; else goto IMPODD_10_49;

   IMPODD_10_45:
    rep=isin_G_H(po,r,39,22,pr);
    if (rep) goto IMPODD_10_50; else return 39;

   IMPODD_10_46:
    rep=isin_G_H(po,r,16,2,pr);
    return rep? 2: 16;

   IMPODD_10_47:
    rep=isin_G_H(po,r,23,14,pr);
    if (rep) goto IMPODD_10_51; else goto IMPODD_10_52;

   IMPODD_10_48:
    rep=isin_G_H(po,r,11,3,pr);
    if (rep) goto IMPODD_10_53; else return 11;

   IMPODD_10_49:
    rep=isin_G_H(po,r,29,23,pr);
    if (rep) goto IMPODD_10_42; else goto IMPODD_10_54;

   IMPODD_10_50:
    rep=isin_G_H(po,r,22,12,pr);
    if (rep) goto IMPODD_10_41; else goto IMPODD_10_55;

   IMPODD_10_51:
    rep=isin_G_H(po,r,14,1,pr);
    return rep? 1: 14;

   IMPODD_10_52:
    rep=isin_G_H(po,r,23,3,pr);
    if (!rep) return 23;
   IMPODD_10_53:
    rep=isin_G_H(po,r,3,2,pr);
    if (rep) return 2; else goto IMPODD_10_57;

   IMPODD_10_54:
    rep=isin_G_H(po,r,29,5,pr);
    if (rep) goto IMPODD_10_58; else return 29;

   IMPODD_10_55:
    rep=isin_G_H(po,r,22,11,pr);
    if (rep) goto IMPODD_10_48;
 /* IMPODD_10_56: */
    rep=isin_G_H(po,r,22,5,pr);
    if (rep) goto IMPODD_10_58; else return 22;

   IMPODD_10_57:
    rep=isin_G_H(po,r,3,1,pr);
    return rep? 1: 3;

   IMPODD_10_58:
    rep=isin_G_H(po,r,5,4,pr);
    if (rep) return 4;
 /* IMPODD_10_59: */
    rep=isin_G_H(po,r,5,3,pr);
    if (rep) goto IMPODD_10_53; else return 5;
  }
}

PERMUTATION
data10(long n1,long n2,GROUP *t)
{
  SID[0]=N;
  switch(n1)
  {
    case 6: if (n2!=2) break;
      *t=initgroup(N,1);
      return _cr(1, 2, 3, 4, 5, 6, 10, 9, 8, 7);
    case 9: if (n2!=3 && n2!=6) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 10, 9, 8, 7);
      return SID;
    case 10: *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 10, 9, 8, 7);
      return SID;
    case 14: case 16:*t=initgroup(N,1);
      return _cr(1, 3, 5, 7, 9, 2, 4, 6, 8, 10);
    case 17: if (n2!=5) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 10, 9, 8, 7);
      return SID;
    case 19: case 20: *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 10, 7, 9);
      return SID;
    case 21: if (n2!=10) break;
      *t=initgroup(N,1);
      return _cr(1, 2, 3, 4, 5, 6, 8, 10, 7, 9);
    case 23: if (n2!=3) break;
      *t=initgroup(N,1);
      return _cr(1, 3, 5, 7, 9, 2, 4, 6, 8, 10);
    case 25: *t=initgroup(N,1);
      return _cr(1, 3, 5, 7, 9, 2, 4, 6, 8, 10);
    case 26: *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 4, 9, 6, 8, 10, 3, 7, 5);
      return _cr(1, 2, 3, 10, 6, 5, 7, 4, 8, 9);
    case 27: if (n2!=17 && n2!=21) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 10, 7, 9);
      return SID;
    case 28: *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 8, 10, 7, 9);
      return SID;
    case 29: if (n2!=5) break;
      *t=initgroup(N,1);
      return _cr(1, 3, 5, 7, 9, 2, 4, 6, 8, 10);
    case 32: *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 4, 9, 6, 8, 10, 3, 7, 5);
      return _cr(1, 2, 3, 10, 6, 5, 7, 4, 8, 9);
    case 36: if (n2!=11) break;
      *t=initgroup(N,1);
      return _cr(1, 3, 5, 7, 9, 2, 4, 6, 8, 10);
    case 38: if (n2!=12) break;
      *t=initgroup(N,1);
      return _cr(1, 3, 5, 7, 9, 2, 4, 6, 8, 10);
    case 39: if (n2!=22) break;
      *t=initgroup(N,1);
      return _cr(1, 3, 5, 7, 9, 2, 4, 6, 8, 10);
    case 40: if (n2!=12) break;
      *t=initgroup(N,1);
      return _cr(1, 2, 3, 4, 5, 6, 7, 8, 10, 9);
    case 41: if (n2!=22 && n2!=40) break;
      *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 7, 8, 10, 9);
  }
  *t=initgroup(N,1); return SID;
}

long
galoismodulo10(GEN pol, GEN dpol, long car)
{
  byteptr d = diffptr;
  GEN ff,p1;
  long av = avma, av1, i,k,p,nbgroup,numer,nbraye,gr[46];
  long lp1,galtyp[NMAX+1],nbremain,j;
  long **TYP10 = partitions(10);
  long **GR10 = (long**) cgeti(45);

  GR10[ 1]= _affl(  4, 1,13,37,42);
  GR10[ 2]= _affl(  3, 13,37,42);
  GR10[ 3]= _affl(  5, 1,13,37,38,42);
  GR10[ 4]= _affl(  4, 13,20,38,42);
  GR10[ 5]= _affl(  7, 1,13,20,21,37,38,42);
  GR10[ 6]= _affl(  5, 1,13,19,37,42);
  GR10[ 7]= _affl( -4, 13,29,38,42);
  GR10[ 8]= _affl( -4, 13,38,40,42);
  GR10[ 9]= _affl(  6, 1,13,19,37,38,42);
  GR10[10]= _affl(  5, 13,19,20,38,42);
  GR10[11]= _affl(  7, 1,10,13,32,37,38,42);
  GR10[12]= _affl(  7, 10,13,20,32,37,38,42);
  GR10[13]= _affl(  7, 9,13,20,29,38,39,42);
  GR10[14]= _affl(  8, 1,13,37,38,39,40,41,42);
  GR10[15]= _affl( -6, 13,21,25,38,40,42);
  GR10[16]= _affl(  7, 13,20,26,37,38,40,42);
  GR10[17]= _affl(  8, 1,13,19,20,21,37,38,42);
  GR10[18]= _affl( -6, 3,13,19,21,38,42);
  GR10[19]= _affl(  7, 1,13,19,21,37,38,42);
  GR10[20]= _affl(  6, 13,19,20,21,38,42);
  GR10[21]= _affl(  9, 1,13,17,19,20,37,38,40,42);
  GR10[22]= _affl( 11, 1,10,13,20,21,30,32,37,38,40,42);
  GR10[23]= _affl( 12, 1,13,20,21,25,26,37,38,39,40,41,42);
  GR10[24]= _affl( -7, 3,13,21,25,38,40,42);
  GR10[25]= _affl(  8, 4,13,20,21,25,38,40,42);
  GR10[26]= _affl( -5, 13,21,29,38,42);
  GR10[27]= _affl( 10, 1,13,17,19,20,21,37,38,40,42);
  GR10[28]= _affl( -8, 3,13,17,19,21,38,40,42);
  GR10[29]= _affl( 14, 1,3,4,13,20,21,25,26,37,38,39,40,41,42);
  GR10[30]= _affl(  8, 1,4,13,21,29,37,38,42);
  GR10[31]= _affl( -6, 3,13,21,29,38,42);
  GR10[32]= _affl(  8, 9,13,20,21,29,38,39,42);
  GR10[33]= _affl( 14, 1,3,13,14,17,19,20,21,26,28,37,38,40,42);
  GR10[34]= _affl( -9, 11,13,21,25,30,32,38,40,42);
  GR10[35]= _affl( 12, 1,3,4,9,13,20,21,29,37,38,39,42);
  GR10[36]= _affl( 18, 1,10,11,12,13,20,21,25,26,30,31,32,37,38,39,40,41,42);
  GR10[37]= _affl(-12, 3,8,11,13,21,25,27,30,32,38,40,42);
  GR10[38]= _affl( 18, 4,8,10,11,13,20,21,22,25,26,28,30,32,37,38,39,40,42);
  GR10[39]= _affl( 24, 1,3,4,8,10,11,12,13,20,21,22,25,26,27,28,30,31,32,37,38,39,40,41,42);
  GR10[40]= _affl( 14, 1,10,13,16,17,19,20,32,34,36,37,38,40,42);
  GR10[41]= _affl( 18, 1,10,13,16,17,19,20,21,23,27,30,32,34,36,37,38,40,42);
  GR10[42]= _affl(-17, 3,8,13,16,17,19,21,23,25,27,30,32,34,36,38,40,42);
  GR10[43]= _affl( 32, 1,3,8,10,13,14,15,16,17,18,19,20,21,23,24,25,26,27,28,30,31,32,33,34,35,36,37,38,39,40,41,42);
  GR10[44]= _affl(-22, 2,3,5,7,8,11,13,16,17,19,21,23,25,27,29,30,32,34,36,38,40,42);

  nbgroup=45; nbremain = car? 12: 33;
  p = 0; for (i=1; i<46; i++) gr[i]=1; 
  for (k=1; k<15; k++, d++)
  {
    p += *d; if (!*d) err(primer1);
    if (smodis(dpol,p)) /* p does not divide dpol */
    {
      av1=avma; ff=simplefactmod(pol,stoi(p));
      p1=(GEN)ff[1]; lp1=lg(p1);
      for (j=1; j<lp1; j++) galtyp[j]=itos((GEN)(p1[lp1-j]));
      for (   ; j<=N ; j++) galtyp[j]=0;
      ranger(galtyp,N); numer=numerotyp(TYP10,galtyp);
      if (car) nbraye = rayergrouppair(GR10,numer,gr,nbgroup);
      else     nbraye = rayergroupimpair(GR10,numer,gr,nbgroup);
      nbremain -= nbraye;
      if (nbremain==1) { avma=av; return car? 44: 45; }
      avma=av1;
    }
  }
  avma=av; return 0;
}
/* BIBLIOTHEQUE POUR LE DEGRE 11 */

long
closure11(GEN po, long ca, long pr)
{
  long nbrac,rep;
  GEN p1,r[NMAX];

  r[0]=p1=roots(po,PRMAX); nbrac=lg(p1)-1;
  if (nbrac!=N) err(talker,"incompatible number of roots in closure11()");
  preci(r,pr);
  if (ca) rep=galoiseven11(po,r,pr); else rep=galoisodd11(po,r,pr);
  return rep;
}

long
galoisodd11(GEN po, GEN *r, long pr)
{
  long rep;
/* ODD_11_1: */
  rep=isin_G_H(po,r,8,4,pr);
  if (!rep) return 8;
/* ODD_11_2: */
  rep=isin_G_H(po,r,4,2,pr);
  return rep? 2: 4;
}

long
galoiseven11(GEN po, GEN *r, long pr)
{
  long rep;
/* EVEN_11_1: */
  rep=isin_G_H(po,r,7,6,pr);
  if (!rep) return 7;
/* EVEN_11_2: */
  rep=isin_G_H(po,r,6,5,pr);
  if (!rep) return 6;
/* EVEN_11_3: */
  rep=isin_G_H(po,r,5,3,pr);
  if (!rep) return 5;
/* EVEN_11_4: */
  rep=isin_G_H(po,r,3,1,pr);
  return rep? 1: 3;
}

PERMUTATION
data11(long n1, GROUP *t)
{
  SID[0]=N;
  switch(n1)
  {
    case 5: *t=initgroup(N,1);
      return _cr(1, 2, 3, 7, 8, 6, 11, 5, 9, 4, 10);
    case 6: *t=initgroup(N,1);
      return _cr(1, 2, 3, 4, 6, 10, 11, 9, 7, 5, 8);
    case 7: *t=initgroup(N,2);
      _aff((*t)[2], 1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 10);
      return SID;
  }
  *t=initgroup(N,1); return SID;
}

static long
rayergrouppair11(long *GR, long num)
{
  long r = 0;

  switch(num)
  {
    case 2: case 3: case 5:
      if (GR[3]) { GR[3]=0; r++; }
    case 6: case 7:
      if (GR[2]) { GR[2]=0; r++; }
    case 4:
      if (GR[1]){ GR[1]=0; r++; }
  }
  return r;
}

static long
rayergroupimpair11(long *GR, long num)
{
  long r = 0;
  if (num > 1 && GR[1]) { GR[1]=0; r++; }
  return r;
}

long
galoismodulo11(GEN pol, GEN dpol, long car)
{
  byteptr d = diffptr;
  long j,k,av = avma,av1,nbgroup,nb,numer,p,lp1,galtyp[NMAX+1];
  char TYPPAIR11[9][12], TYPIMPAIR11[6][12], **T;
  char GROUPPAIR11[6] = { 1,1,1,1,1,1 }, GROUPIMPAIR11[4] = { 1,1,1,1 };
  GEN ff,p1;

  T = (char**) TYPIMPAIR11; T[0][0] = 5; 
  _aff(T[1], 11, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  _aff(T[2], 10, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  _aff(T[3],  5, 5, 1, 0, 0, 0, 0, 0, 0, 0, 0);
  _aff(T[4],  2, 2, 2, 2, 2, 1, 0, 0, 0, 0, 0);
  _aff(T[5],  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1);

  T = (char**) TYPPAIR11; T[0][0] = 8;
  _aff(T[1], 11, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  _aff(T[2],  8, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0);
  _aff(T[3],  6, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0);
  _aff(T[4],  5, 5, 1, 0, 0, 0, 0, 0, 0, 0, 0);
  _aff(T[5],  4, 4, 1, 1, 1, 0, 0, 0, 0, 0, 0);
  _aff(T[6],  3, 3, 3, 1, 1, 0, 0, 0, 0, 0, 0);
  _aff(T[7],  2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0);
  _aff(T[8],  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1);

  p = 0;
  if (car)
  {
    nbgroup=5;
    for (k=1; k<12; k++, d++)
    {
      p += *d; if (!*d) err(primer1);
      if (smodis(dpol,p)) /* p does not divide dpol */
      {
        av1=avma; ff=simplefactmod(pol,stoi(p));
        p1=(GEN)ff[1]; lp1=lg(p1);
        for (j=1; j<lp1; j++) galtyp[j]=itos((GEN)(p1[lp1-j]));
        for (   ; j<=N ; j++) galtyp[j]=0;
        ranger(galtyp,N); numer=numerotyp((long**)TYPPAIR11,galtyp);
        if (!numer) { avma=av; return 1; }
        nb=rayergrouppair11((long*)GROUPPAIR11,numer); nbgroup -= nb;
        if (nbgroup==1) { avma=av; return 1; }
        avma=av1;
      }
    }
  }
  else
  {
    nbgroup=3;
    for (k=1; k<12; k++, d++)
    {
      p += *d; if (!*d) err(primer1);
      if (smodis(dpol,p)) /* p does not divide dpol */
      {
        av1=avma; ff=simplefactmod(pol,stoi(p));
        p1=(GEN)ff[1]; lp1=lg(p1);
        for (j=1; j<lp1; j++) galtyp[j]=itos((GEN)(p1[lp1-j]));
        for (   ; j<=N ; j++) galtyp[j]=0;
        ranger(galtyp,N); numer=numerotyp((long**)TYPIMPAIR11,galtyp);
        if (!numer) { avma=av; return 1; }
        nb=rayergroupimpair11((long*)GROUPIMPAIR11,numer); nbgroup -= nb;
        if (nbgroup==1) { avma=av; return 1; }
        avma=av1;
      }
    }
  }
  avma=av; return 0;
}
